# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

_stub_error = NotImplementedError("This is a stub.")

# mem fence
CLK_LOCAL_MEM_FENCE = 0x1
CLK_GLOBAL_MEM_FENCE = 0x2


def get_global_id(*args, **kargs):
    """
    OpenCL get_global_id()
    """
    raise _stub_error


def get_local_id(*args, **kargs):
    """
    OpenCL get_local_id()
    """
    raise _stub_error


def get_global_size(*args, **kargs):
    """
    OpenCL get_global_size()
    """
    raise _stub_error


def get_local_size(*args, **kargs):
    """
    OpenCL get_local_size()
    """
    raise _stub_error


def get_group_id(*args, **kargs):
    """
    OpenCL get_group_id()
    """
    raise _stub_error


def get_num_groups(*args, **kargs):
    """
    OpenCL get_num_groups()
    """
    raise _stub_error


def get_work_dim(*args, **kargs):
    """
    OpenCL get_work_dim()
    """
    raise _stub_error


def barrier(*args, **kargs):
    """
    OpenCL barrier()
    """
    raise _stub_error


def mem_fence(*args, **kargs):
    """
    OpenCL mem_fence()
    """
    raise _stub_error


def sub_group_barrier():
    """
    OpenCL 2.0 sub_group_barrier
    """
    raise _stub_error


class Stub(object):
    """A stub object to represent special objects which is meaningless
    outside the context of DPPY compilation context.
    """

    _description_ = "<dppy special value>"
    __slots__ = ()  # don't allocate __dict__

    def __new__(cls):
        raise NotImplementedError("%s is not instantiable" % cls)

    def __repr__(self):
        return self._description_


# -------------------------------------------------------------------------------
# local memory


class local(Stub):
    """local namespace"""

    _description_ = "<local>"

    def array(shape, dtype):
        """local.array(shape, dtype)

        Allocate a local array.
        """


# -------------------------------------------------------------------------------
# atomic


class atomic(Stub):
    """atomic namespace"""

    _description_ = "<atomic>"

    def add():
        """add(ary, idx, val)

        Perform atomic ary[idx] += val.

        Returns the old value at the index location as if it is loaded atomically.

        .. note:: Supported on int32, int64, float32, float64 operands only.
        """

    def sub():
        """sub(ary, idx, val)

        Perform atomic ary[idx] -= val.

        Returns the old value at the index location as if it is loaded atomically.

        .. note:: Supported on int32, int64, float32, float64 operands only.
        """



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from numba import types
from numba.core.extending import overload, register_jitable
from numba.core.typing import signature

import numba_dppy
import numba_dppy.dpnp_glue as dpnp_lowering
import numba_dppy.dpnp_glue.dpnpimpl as dpnp_ext
from numba_dppy import dpctl_functions

from . import stubs


@register_jitable
def common_impl(a, out, dpnp_func, print_debug):
    if a.size == 0:
        raise ValueError("Passed Empty array")

    sycl_queue = dpctl_functions.get_current_queue()
    a_usm = dpctl_functions.malloc_shared(a.size * a.itemsize, sycl_queue)
    event = dpctl_functions.queue_memcpy(
        sycl_queue, a_usm, a.ctypes, a.size * a.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    out_usm = dpctl_functions.malloc_shared(a.itemsize, sycl_queue)

    axes, axes_ndim = 0, 0
    initial = 0
    where = 0

    dpnp_func(out_usm, a_usm, a.shapeptr, a.ndim, axes, axes_ndim, initial, where)

    event = dpctl_functions.queue_memcpy(
        sycl_queue, out.ctypes, out_usm, out.size * out.itemsize
    )
    dpctl_functions.event_wait(event)
    dpctl_functions.event_delete(event)

    dpctl_functions.free_with_queue(a_usm, sycl_queue)
    dpctl_functions.free_with_queue(out_usm, sycl_queue)

    dpnp_ext._dummy_liveness_func([a.size, out.size])

    if print_debug:
        print("dpnp implementation")


@overload(stubs.dpnp.sum)
def dpnp_sum_impl(a):
    name = "sum"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.6.1dev/dpnp/backend/kernels/dpnp_krnl_reduction.cpp#L59

    Function declaration:
    void dpnp_sum_c(void* result_out,
                    const void* input_in,
                    const size_t* input_shape,
                    const size_t input_shape_ndim,
                    const long* axes,
                    const size_t axes_ndim,
                    const void* initial,
                    const long* where)

    """
    sig = signature(
        ret_type,
        types.voidptr,  # void* result_out,
        types.voidptr,  # const void* input_in,
        types.voidptr,  # const size_t* input_shape,
        types.intp,  # const size_t input_shape_ndim,
        types.voidptr,  # const long* axes,
        types.intp,  # const size_t axes_ndim,
        types.voidptr,  # const void* initial,
        types.voidptr,  # const long* where)
    )
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [a.dtype.name, "NONE"], sig)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a):
        out = np.empty(1, dtype=a.dtype)
        common_impl(a, out, dpnp_func, PRINT_DEBUG)

        return out[0]

    return dpnp_impl


@overload(stubs.dpnp.prod)
def dpnp_prod_impl(a):
    name = "prod"
    dpnp_lowering.ensure_dpnp(name)

    ret_type = types.void
    """
    dpnp source:
    https://github.com/IntelPython/dpnp/blob/0.6.1dev/dpnp/backend/kernels/dpnp_krnl_reduction.cpp#L129

    Function declaration:
    void dpnp_prod_c(void* result_out,
                     const void* input_in,
                     const size_t* input_shape,
                     const size_t input_shape_ndim,
                     const long* axes,
                     const size_t axes_ndim,
                     const void* initial, // type must be _DataType_output
                     const long* where)
    """
    sig = signature(
        ret_type,
        types.voidptr,  # void* result_out,
        types.voidptr,  # const void* input_in,
        types.voidptr,  # const size_t* input_shape,
        types.intp,  # const size_t input_shape_ndim,
        types.voidptr,  # const long* axes,
        types.intp,  # const long* axes,
        types.voidptr,  # const void* initial, // type must be _DataType_output
        types.voidptr,  # const long* where)
    )
    dpnp_func = dpnp_ext.dpnp_func("dpnp_" + name, [a.dtype.name, "NONE"], sig)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a):
        out = np.empty(1, dtype=a.dtype)

        common_impl(a, out, dpnp_func, PRINT_DEBUG)
        return out[0]

    return dpnp_impl


@overload(stubs.dpnp.nansum)
def dpnp_nansum_impl(a):
    name = "nansum"
    dpnp_lowering.ensure_dpnp(name)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a):
        a_ravel = a.ravel()
        a_ravel_copy = np.copy(a_ravel)

        for i in range(len(a_ravel_copy)):
            if np.isnan(a_ravel_copy[i]):
                a_ravel_copy[i] = 0

        result = numba_dppy.dpnp.sum(a_ravel_copy)
        dpnp_ext._dummy_liveness_func([a.size, a_ravel_copy.size])

        if PRINT_DEBUG:
            print("dpnp implementation")

        return result

    return dpnp_impl


@overload(stubs.dpnp.nanprod)
def dpnp_nanprod_impl(a):
    name = "nanprod"
    dpnp_lowering.ensure_dpnp(name)

    PRINT_DEBUG = dpnp_lowering.DEBUG

    def dpnp_impl(a):
        a_ravel = a.ravel()
        a_ravel_copy = np.copy(a_ravel)

        for i in range(len(a_ravel_copy)):
            if np.isnan(a_ravel_copy[i]):
                a_ravel_copy[i] = 1

        result = numba_dppy.dpnp.prod(a_ravel_copy)
        dpnp_ext._dummy_liveness_func([a.size, a_ravel_copy.size])

        if PRINT_DEBUG:
            print("dpnp implementation")

        return result

    return dpnp_impl



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import dpctl
import numpy as np

import numba_dppy as dppy

from . import _helper


@unittest.skipUnless(_helper.has_gpu_queues(), "test only on GPU system")
class TestDPPYFunc(unittest.TestCase):
    N = 257

    def test_dppy_func_device_array(self):
        @dppy.func
        def g(a):
            return a + 1

        @dppy.kernel
        def f(a, b):
            i = dppy.get_global_id(0)
            b[i] = g(a[i])

        a = np.ones(self.N)
        b = np.ones(self.N)

        device = dpctl.SyclDevice("opencl:gpu")
        with dpctl.device_context(device):
            f[self.N, dppy.DEFAULT_LOCAL_SIZE](a, b)

        self.assertTrue(np.all(b == 2))

    def test_dppy_func_ndarray(self):
        @dppy.func
        def g(a):
            return a + 1

        @dppy.kernel
        def f(a, b):
            i = dppy.get_global_id(0)
            b[i] = g(a[i])

        @dppy.kernel
        def h(a, b):
            i = dppy.get_global_id(0)
            b[i] = g(a[i]) + 1

        a = np.ones(self.N)
        b = np.ones(self.N)

        device = dpctl.SyclDevice("opencl:gpu")
        with dpctl.device_context(device):
            f[self.N, dppy.DEFAULT_LOCAL_SIZE](a, b)

            self.assertTrue(np.all(b == 2))

            h[self.N, dppy.DEFAULT_LOCAL_SIZE](a, b)

            self.assertTrue(np.all(b == 3))


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from functools import singledispatch

import llvmlite.llvmpy.core as lc
from numba.core import cgutils, types, typing
from numba.core.imputils import Registry

from numba_dppy.utils import address_space

registry = Registry()
lower = registry.lower


def declare_print(lmod):
    voidptrty = lc.Type.pointer(lc.Type.int(8), addrspace=address_space.GENERIC)
    printfty = lc.Type.function(lc.Type.int(), [voidptrty], var_arg=True)
    printf = cgutils.get_or_insert_function(lmod, printfty, "printf")
    return printf


@singledispatch
def print_item(ty, context, builder, val):
    """
    Handle printing of a single value of the given Numba type.
    A (format string, [list of arguments]) is returned that will allow
    forming the final printf()-like call.
    """
    raise NotImplementedError("printing unimplemented for values of type %s" % (ty,))


@print_item.register(types.Integer)
@print_item.register(types.IntegerLiteral)
def int_print_impl(ty, context, builder, val):
    if ty in types.unsigned_domain:
        rawfmt = "%llu"
        dsttype = types.uint64
    else:
        rawfmt = "%lld"
        dsttype = types.int64
    fmt = context.insert_const_string(builder.module, rawfmt)  # noqa
    lld = context.cast(builder, val, ty, dsttype)
    return rawfmt, [lld]


@print_item.register(types.Float)
def real_print_impl(ty, context, builder, val):
    lld = context.cast(builder, val, ty, types.float64)
    return "%f", [lld]


@print_item.register(types.StringLiteral)
def const_print_impl(ty, context, builder, sigval):
    pyval = ty.literal_value
    assert isinstance(pyval, str)  # Ensured by lowering
    rawfmt = "%s"
    val = context.insert_const_string(builder.module, pyval)
    return rawfmt, [val]


@lower(print, types.VarArg(types.Any))
def print_varargs(context, builder, sig, args):
    """This function is a generic 'print' wrapper for arbitrary types.
    It dispatches to the appropriate 'print' implementations above
    depending on the detected real types in the signature."""

    formats = []
    values = []

    only_str = True
    for i, (argtype, argval) in enumerate(zip(sig.args, args)):
        argfmt, argvals = print_item(argtype, context, builder, argval)
        formats.append(argfmt)
        values.extend(argvals)
        if argfmt != "%s":
            only_str = False

    if only_str:
        raise ValueError("We do not support printing string alone!")

    rawfmt = " ".join(formats) + "\n"
    fmt = context.insert_const_string(builder.module, rawfmt)

    va_arg = [fmt]
    va_arg.extend(values)
    va_arg = tuple(va_arg)

    dppy_print = declare_print(builder.module)

    builder.call(dppy_print, va_arg)

    return context.get_dummy_value()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import dpctl
import numpy as np
from numba import njit, prange
from numba.tests.support import captured_stdout

import numba_dppy as dppy
from numba_dppy import config as dppy_config

from . import _helper


@unittest.skipUnless(_helper.has_gpu_queues(), "test only on GPU system")
class TestOffloadDiagnostics(unittest.TestCase):
    def test_parfor(self):
        def prange_func():
            n = 10
            a = np.ones((n), dtype=np.float64)
            b = np.ones((n), dtype=np.float64)
            c = np.ones((n), dtype=np.float64)
            for i in prange(n // 2):
                a[i] = b[i] + c[i]

            return a

        device = dpctl.SyclDevice("opencl:gpu")
        with dpctl.device_context(device):
            dppy_config.OFFLOAD_DIAGNOSTICS = 1
            jitted = njit(parallel=True)(prange_func)

            with captured_stdout() as got:
                jitted()

            dppy_config.OFFLOAD_DIAGNOSTICS = 0
            self.assertTrue("Auto-offloading" in got.getvalue())
            self.assertTrue("Device -" in got.getvalue())

    def test_kernel(self):
        @dppy.kernel
        def parallel_sum(a, b, c):
            i = dppy.get_global_id(0)
            c[i] = a[i] + b[i]

        global_size = 10
        N = global_size

        a = np.array(np.random.random(N), dtype=np.float32)
        b = np.array(np.random.random(N), dtype=np.float32)
        c = np.ones_like(a)

        device = dpctl.SyclDevice("opencl:gpu")
        with dpctl.device_context(device):
            dppy_config.OFFLOAD_DIAGNOSTICS = 1

            with captured_stdout() as got:
                parallel_sum[global_size, dppy.DEFAULT_LOCAL_SIZE](a, b, c)

            dppy_config.OFFLOAD_DIAGNOSTICS = 0
            self.assertTrue("Auto-offloading" in got.getvalue())
            self.assertTrue("Device -" in got.getvalue())


if __name__ == "__main__":
    unittest.main()



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import contextlib
import unittest

import dpctl
from numba.tests.support import captured_stdout

from numba_dppy import config


def has_gpu_queues(backend="opencl"):
    """
    Checks if dpctl is able to select a GPU device that defaults to
    an OpenCL GPU.
    """
    return bool(dpctl.get_num_devices(backend=backend, device_type="gpu"))


def has_cpu_queues(backend="opencl"):
    """
    Checks if dpctl is able to select a CPU device that defaults to
    an OpenCL CPU.
    """
    return bool(dpctl.get_num_devices(backend=backend, device_type="cpu"))


def has_sycl_platforms():
    """
    Checks if dpctl is able to identify a non-host SYCL platform.
    """
    platforms = dpctl.get_platforms()
    for p in platforms:
        if p.backend is not dpctl.backend_type.host:
            return True
    return False


def is_gen12(device_type):
    with dpctl.device_context(device_type):
        q = dpctl.get_current_queue()
        device = q.get_sycl_device()
        name = device.name
        if "Gen12" in name:
            return True

        return False


def platform_not_supported(device_type):
    import platform

    platform = platform.system()
    device = device_type.split(":")[0]

    if device == "level_zero" and platform == "Windows":
        return True

    return False


def skip_test(device_type):
    skip = False
    try:
        with dpctl.device_context(device_type):
            pass
    except Exception:
        skip = True

    if not skip:
        if platform_not_supported(device_type):
            skip = True

    return skip


@contextlib.contextmanager
def override_config(name, value, config=config):
    """
    Extends `numba/tests/support.py:override_config()` with argument `config`
    which is `numba_dppy.config` by default.
    """
    old_value = getattr(config, name)
    setattr(config, name, value)
    try:
        yield
    finally:
        setattr(config, name, old_value)


def _id(obj):
    return obj


def expectedFailureIf(condition):
    """
    Expected failure for a test if the condition is true.
    """
    if condition:
        return unittest.expectedFailure
    return _id


def ensure_dpnp():
    try:
        from numba_dppy.dpnp_glue import dpnp_fptr_interface as dpnp_glue

        return True
    except:
        return False


@contextlib.contextmanager
def dpnp_debug():
    import numba_dppy.dpnp_glue as dpnp_lowering

    old, dpnp_lowering.DEBUG = dpnp_lowering.DEBUG, 1
    yield
    dpnp_lowering.DEBUG = old


@contextlib.contextmanager
def assert_dpnp_implementaion():
    from numba.tests.support import captured_stdout

    with captured_stdout() as stdout, dpnp_debug():
        yield

    assert "dpnp implementation" in stdout.getvalue(), "dpnp implementation is not used"


@contextlib.contextmanager
def assert_auto_offloading(parfor_offloaded=1, parfor_offloaded_failure=0):
    """
    If ``parfor_offloaded`` is not provided this context_manager
    will check for 1 occurrance of success message. Developers
    can always specify how many parfor offload success message
    is expected.
    If ``parfor_offloaded_failure`` is not provided the default
    behavior is to expect 0 failure message, in other words, we
    expect all parfors present in the code to be successfully
    offloaded to GPU.
    """
    old_debug = config.DEBUG
    config.DEBUG = 1

    with captured_stdout() as stdout:
        yield

    config.DEBUG = old_debug

    got_parfor_offloaded = stdout.getvalue().count("Parfor offloaded to")
    assert parfor_offloaded == got_parfor_offloaded, (
        "Expected %d parfor(s) to be auto offloaded, instead got %d parfor(s) auto offloaded"
        % (parfor_offloaded, got_parfor_offloaded)
    )

    got_parfor_offloaded_failure = stdout.getvalue().count(
        "Failed to offload parfor to"
    )
    assert parfor_offloaded_failure == got_parfor_offloaded_failure, (
        "Expected %d parfor(s) to be not auto offloaded, instead got %d parfor(s) not auto offloaded"
        % (parfor_offloaded_failure, got_parfor_offloaded_failure)
    )



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import copy
from collections import OrderedDict
import linecache
import os
import sys
import numpy as np

import numba
from numba.core import compiler, ir, types, sigutils, lowering, funcdesc
from numba.parfors import parfor
from numba.parfors.parfor_lowering import _lower_parfor_parallel
import numba_dppy, numba_dppy as dppy
from numba.core.ir_utils import (
    add_offset_to_labels,
    replace_var_names,
    remove_dels,
    legalize_names,
    mk_unique_var,
    rename_labels,
    get_name_var_table,
    visit_vars_inner,
    guard,
    find_callname,
    remove_dead,
    get_call_table,
    is_pure,
    build_definitions,
    get_np_ufunc_typ,
    get_unused_var_name,
    find_potential_aliases,
    is_const_call,
)

from numba.core.typing import signature

import warnings
from numba.core.errors import NumbaParallelSafetyWarning, NumbaPerformanceWarning

from .dufunc_inliner import dufunc_inliner
from numba_dppy.driver import KernelLaunchOps
import dpctl
from numba_dppy import config
from numba_dppy.target import DPPYTargetContext
from numba_dppy.dppy_array_type import DPPYArray
from numba_dppy.utils import address_space, npytypes_array_to_dppy_array


def _print_block(block):
    for i, inst in enumerate(block.body):
        print("    ", i, inst)


def _print_body(body_dict):
    """Pretty-print a set of IR blocks."""
    for label, block in body_dict.items():
        print("label: ", label)
        _print_block(block)


# This loop scheduler is pretty basic, there is only
# 3 dimension allowed in OpenCL, so to make the backend
# functional we will schedule the first 3 dimensions
# through OpenCL and generate for loops for the remaining
# dimensions
def _schedule_loop(parfor_dim, legal_loop_indices, loop_ranges, param_dict):
    gufunc_txt = ""
    global_id_dim = 0
    for_loop_dim = parfor_dim

    if parfor_dim > 3:
        global_id_dim = 3
    else:
        global_id_dim = parfor_dim

    for eachdim in range(global_id_dim):
        gufunc_txt += (
            "    "
            + legal_loop_indices[eachdim]
            + " = "
            + "dppy.get_global_id("
            + str(eachdim)
            + ")\n"
        )

    for eachdim in range(global_id_dim, for_loop_dim):
        for indent in range(1 + (eachdim - global_id_dim)):
            gufunc_txt += "    "

        start, stop, step = loop_ranges[eachdim]
        start = param_dict.get(str(start), start)
        stop = param_dict.get(str(stop), stop)
        gufunc_txt += (
            "for "
            + legal_loop_indices[eachdim]
            + " in range("
            + str(start)
            + ", "
            + str(stop)
            + " + 1):\n"
        )

    for eachdim in range(global_id_dim, for_loop_dim):
        for indent in range(1 + (eachdim - global_id_dim)):
            gufunc_txt += "    "

    return gufunc_txt


def _dbgprint_after_each_array_assignments(lowerer, loop_body, typemap):
    for label, block in loop_body.items():
        new_block = block.copy()
        new_block.clear()
        loc = block.loc
        scope = block.scope
        for inst in block.body:
            new_block.append(inst)
            # Append print after assignment
            if isinstance(inst, ir.Assign):
                # Only apply to numbers
                if typemap[inst.target.name] not in types.number_domain:
                    continue

                # Make constant string
                strval = "{} =".format(inst.target.name)
                strconsttyp = types.StringLiteral(strval)

                lhs = ir.Var(scope, mk_unique_var("str_const"), loc)
                assign_lhs = ir.Assign(
                    value=ir.Const(value=strval, loc=loc), target=lhs, loc=loc
                )
                typemap[lhs.name] = strconsttyp
                new_block.append(assign_lhs)

                # Make print node
                print_node = ir.Print(args=[lhs, inst.target], vararg=None, loc=loc)
                new_block.append(print_node)
                sig = numba.typing.signature(
                    types.none, typemap[lhs.name], typemap[inst.target.name]
                )
                lowerer.fndesc.calltypes[print_node] = sig
        loop_body[label] = new_block


def replace_var_with_array_in_block(vars, block, typemap, calltypes):
    new_block = []
    for inst in block.body:
        if isinstance(inst, ir.Assign) and inst.target.name in vars:
            const_node = ir.Const(0, inst.loc)
            const_var = ir.Var(
                inst.target.scope, mk_unique_var("$const_ind_0"), inst.loc
            )
            typemap[const_var.name] = types.uintp
            const_assign = ir.Assign(const_node, const_var, inst.loc)
            new_block.append(const_assign)

            setitem_node = ir.SetItem(inst.target, const_var, inst.value, inst.loc)
            calltypes[setitem_node] = signature(
                types.none,
                types.npytypes.Array(typemap[inst.target.name], 1, "C"),
                types.intp,
                typemap[inst.target.name],
            )
            new_block.append(setitem_node)
            continue
        elif isinstance(inst, parfor.Parfor):
            replace_var_with_array_internal(
                vars, {0: inst.init_block}, typemap, calltypes
            )
            replace_var_with_array_internal(vars, inst.loop_body, typemap, calltypes)

        new_block.append(inst)
    return new_block


def replace_var_with_array_internal(vars, loop_body, typemap, calltypes):
    for label, block in loop_body.items():
        block.body = replace_var_with_array_in_block(vars, block, typemap, calltypes)


def replace_var_with_array(vars, loop_body, typemap, calltypes):
    replace_var_with_array_internal(vars, loop_body, typemap, calltypes)
    for v in vars:
        el_typ = typemap[v]
        typemap.pop(v, None)
        typemap[v] = types.npytypes.Array(el_typ, 1, "C")


def wrap_loop_body(loop_body):
    blocks = loop_body.copy()  # shallow copy is enough
    first_label = min(blocks.keys())
    last_label = max(blocks.keys())
    loc = blocks[last_label].loc
    blocks[last_label].body.append(ir.Jump(first_label, loc))
    return blocks


def unwrap_loop_body(loop_body):
    last_label = max(loop_body.keys())
    loop_body[last_label].body = loop_body[last_label].body[:-1]


def legalize_names_with_typemap(names, typemap):
    """We use ir_utils.legalize_names to replace internal IR variable names
    containing illegal characters (e.g. period) with a legal character
    (underscore) so as to create legal variable names.
    The original variable names are in the typemap so we also
    need to add the legalized name to the typemap as well.
    """
    outdict = legalize_names(names)
    # For each pair in the dict of legalized names...
    for x, y in outdict.items():
        # If the name had some legalization change to it...
        if x != y:
            # Set the type of the new name the same as the type of the old name.
            typemap[y] = typemap[x]
    return outdict


def to_scalar_from_0d(x):
    if isinstance(x, types.ArrayCompatible):
        if x.ndim == 0:
            return x.dtype
    return x


def find_setitems_block(setitems, block, typemap):
    for inst in block.body:
        if isinstance(inst, ir.StaticSetItem) or isinstance(inst, ir.SetItem):
            setitems.add(inst.target.name)
        elif isinstance(inst, parfor.Parfor):
            find_setitems_block(setitems, inst.init_block, typemap)
            find_setitems_body(setitems, inst.loop_body, typemap)


def find_setitems_body(setitems, loop_body, typemap):
    """
    Find the arrays that are written into (goes into setitems)
    """
    for label, block in loop_body.items():
        find_setitems_block(setitems, block, typemap)


def _create_gufunc_for_regular_parfor():
    # TODO
    pass


def _create_gufunc_for_reduction_parfor():
    raise ValueError("Reductions are not yet supported via parfor")


def _create_gufunc_for_parfor_body(
    lowerer,
    parfor,
    typemap,
    typingctx,
    targetctx,
    flags,
    loop_ranges,
    locals,
    has_aliases,
    index_var_typ,
    races,
):
    """
    Takes a parfor and creates a gufunc function for its body. There
    are two parts to this function:

        1) Code to iterate across the iteration space as defined by
           the schedule.
        2) The parfor body that does the work for a single point in
           the iteration space.

    Part 1 is created as Python text for simplicity with a sentinel
    assignment to mark the point in the IR where the parfor body
    should be added. This Python text is 'exec'ed into existence and its
    IR retrieved with run_frontend. The IR is scanned for the sentinel
    assignment where that basic block is split and the IR for the parfor
    body inserted.
    """

    loc = parfor.init_block.loc

    # The parfor body and the main function body share ir.Var nodes.
    # We have to do some replacements of Var names in the parfor body
    # to make them legal parameter names. If we don't copy then the
    # Vars in the main function also would incorrectly change their name.

    loop_body = copy.copy(parfor.loop_body)
    remove_dels(loop_body)

    parfor_dim = len(parfor.loop_nests)
    loop_indices = [l.index_variable.name for l in parfor.loop_nests]

    # Get all the parfor params.
    parfor_params = parfor.params

    for start, stop, step in loop_ranges:
        if isinstance(start, ir.Var):
            parfor_params.add(start.name)
        if isinstance(stop, ir.Var):
            parfor_params.add(stop.name)

    # Get just the outputs of the parfor.
    parfor_outputs = numba.parfors.parfor.get_parfor_outputs(parfor, parfor_params)

    # Get all parfor reduction vars, and operators.
    typemap = lowerer.fndesc.typemap

    parfor_redvars, parfor_reddict = numba.parfors.parfor.get_parfor_reductions(
        lowerer.func_ir, parfor, parfor_params, lowerer.fndesc.calltypes
    )
    has_reduction = False if len(parfor_redvars) == 0 else True

    if has_reduction:
        _create_gufunc_for_reduction_parfor()

    # Compute just the parfor inputs as a set difference.
    parfor_inputs = sorted(list(set(parfor_params) - set(parfor_outputs)))

    for race in races:
        msg = (
            "Variable %s used in parallel loop may be written "
            "to simultaneously by multiple workers and may result "
            "in non-deterministic or unintended results." % race
        )
        warnings.warn(NumbaParallelSafetyWarning(msg, loc))
    replace_var_with_array(races, loop_body, typemap, lowerer.fndesc.calltypes)

    if config.DEBUG_ARRAY_OPT >= 1:
        print("parfor_params = ", parfor_params, type(parfor_params))
        print("parfor_outputs = ", parfor_outputs, type(parfor_outputs))
        print("parfor_inputs = ", parfor_inputs, type(parfor_inputs))

    # Reorder all the params so that inputs go first then outputs.
    parfor_params = parfor_inputs + parfor_outputs

    def addrspace_from(params, def_addr):
        addrspaces = []
        for p in params:
            if isinstance(to_scalar_from_0d(typemap[p]), types.npytypes.Array):
                addrspaces.append(def_addr)
            else:
                addrspaces.append(None)
        return addrspaces

    addrspaces = addrspace_from(parfor_params, address_space.GLOBAL)

    if config.DEBUG_ARRAY_OPT >= 1:
        print("parfor_params = ", parfor_params, type(parfor_params))
        print("loop_indices = ", loop_indices, type(loop_indices))
        print("loop_body = ", loop_body, type(loop_body))
        _print_body(loop_body)

    # Some Var are not legal parameter names so create a dict of
    # potentially illegal param name to guaranteed legal name.
    param_dict = legalize_names_with_typemap(parfor_params, typemap)
    if config.DEBUG_ARRAY_OPT >= 1:
        print("param_dict = ", sorted(param_dict.items()), type(param_dict))

    # Some loop_indices are not legal parameter names so create a dict
    # of potentially illegal loop index to guaranteed legal name.
    ind_dict = legalize_names_with_typemap(loop_indices, typemap)
    # Compute a new list of legal loop index names.
    legal_loop_indices = [ind_dict[v] for v in loop_indices]

    if config.DEBUG_ARRAY_OPT >= 1:
        print("ind_dict = ", sorted(ind_dict.items()), type(ind_dict))
        print("legal_loop_indices = ", legal_loop_indices, type(legal_loop_indices))

        for pd in parfor_params:
            print("pd = ", pd)
            print("pd type = ", typemap[pd], type(typemap[pd]))

    # Get the types of each parameter.
    param_types = [to_scalar_from_0d(typemap[v]) for v in parfor_params]

    param_types_addrspaces = copy.copy(param_types)

    # Calculate types of args passed to gufunc.
    func_arg_types = [typemap[v] for v in (parfor_inputs + parfor_outputs)]
    assert len(param_types_addrspaces) == len(addrspaces)
    for i in range(len(param_types_addrspaces)):
        if addrspaces[i] is not None:
            # Convert Numba's npytype.Array to DPPYArray data type. DPPYArray
            # allows us to specify an address space for the data and other
            # pointer arguments for the array.
            param_types_addrspaces[i] = npytypes_array_to_dppy_array(
                param_types_addrspaces[i], addrspaces[i]
            )

    def print_arg_with_addrspaces(args):
        for a in args:
            print(a, type(a))
            if isinstance(a, types.npytypes.Array):
                print("addrspace:", a.addrspace)

    if config.DEBUG_ARRAY_OPT >= 1:
        print_arg_with_addrspaces(param_types)
        print("func_arg_types = ", func_arg_types, type(func_arg_types))

    # Replace illegal parameter names in the loop body with legal ones.
    replace_var_names(loop_body, param_dict)
    # remember the name before legalizing as the actual arguments
    parfor_args = parfor_params
    # Change parfor_params to be legal names.
    parfor_params = [param_dict[v] for v in parfor_params]
    parfor_params_orig = parfor_params

    parfor_params = []
    ascontig = False
    for pindex in range(len(parfor_params_orig)):
        if (
            ascontig
            and pindex < len(parfor_inputs)
            and isinstance(param_types[pindex], types.npytypes.Array)
        ):
            parfor_params.append(parfor_params_orig[pindex] + "param")
        else:
            parfor_params.append(parfor_params_orig[pindex])

    # Change parfor body to replace illegal loop index vars with legal ones.
    replace_var_names(loop_body, ind_dict)
    loop_body_var_table = get_name_var_table(loop_body)
    sentinel_name = get_unused_var_name("__sentinel__", loop_body_var_table)

    if config.DEBUG_ARRAY_OPT >= 1:
        print("legal parfor_params = ", parfor_params, type(parfor_params))

    # Determine the unique names of the scheduling and gufunc functions.
    gufunc_name = "__numba_parfor_gufunc_%s" % (parfor.id)

    if config.DEBUG_ARRAY_OPT:
        # print("sched_func_name ", type(sched_func_name), sched_func_name)
        print("gufunc_name ", type(gufunc_name), gufunc_name)

    gufunc_txt = ""

    # Create the gufunc function.
    gufunc_txt += "def " + gufunc_name
    gufunc_txt += "(" + (", ".join(parfor_params)) + "):\n"

    gufunc_txt += _schedule_loop(
        parfor_dim, legal_loop_indices, loop_ranges, param_dict
    )

    # Add the sentinel assignment so that we can find the loop body position
    # in the IR.
    gufunc_txt += "    "
    gufunc_txt += sentinel_name + " = 0\n"

    # gufunc returns nothing
    gufunc_txt += "    return None\n"

    if config.DEBUG_ARRAY_OPT:
        print("gufunc_txt = ", type(gufunc_txt), "\n", gufunc_txt)
        sys.stdout.flush()
    # Force gufunc outline into existence.
    globls = {"np": np, "numba": numba, "dppy": dppy}
    locls = {}
    exec(gufunc_txt, globls, locls)
    gufunc_func = locls[gufunc_name]

    if config.DEBUG_ARRAY_OPT:
        print("gufunc_func = ", type(gufunc_func), "\n", gufunc_func)
    # Get the IR for the gufunc outline.
    gufunc_ir = compiler.run_frontend(gufunc_func)

    if config.DEBUG_ARRAY_OPT:
        print("gufunc_ir dump ", type(gufunc_ir))
        gufunc_ir.dump()
        print("loop_body dump ", type(loop_body))
        _print_body(loop_body)

    # rename all variables in gufunc_ir afresh
    var_table = get_name_var_table(gufunc_ir.blocks)
    new_var_dict = {}
    reserved_names = [sentinel_name] + list(param_dict.values()) + legal_loop_indices
    for name, var in var_table.items():
        if not (name in reserved_names):
            new_var_dict[name] = mk_unique_var(name)
    replace_var_names(gufunc_ir.blocks, new_var_dict)
    if config.DEBUG_ARRAY_OPT:
        print("gufunc_ir dump after renaming ")
        gufunc_ir.dump()

    prs_dict = {}
    pss_dict = {}
    pspmd_dict = {}

    gufunc_param_types = param_types

    if config.DEBUG_ARRAY_OPT:
        print(
            "gufunc_param_types = ", type(gufunc_param_types), "\n", gufunc_param_types
        )

    gufunc_stub_last_label = max(gufunc_ir.blocks.keys()) + 1

    # Add gufunc stub last label to each parfor.loop_body label to prevent
    # label conflicts.
    loop_body = add_offset_to_labels(loop_body, gufunc_stub_last_label)
    # new label for splitting sentinel block
    new_label = max(loop_body.keys()) + 1

    # If enabled, add a print statement after every assignment.
    if config.DEBUG_ARRAY_OPT_RUNTIME:
        _dbgprint_after_each_array_assignments(lowerer, loop_body, typemap)

    if config.DEBUG_ARRAY_OPT:
        print("parfor loop body")
        _print_body(loop_body)

    wrapped_blocks = wrap_loop_body(loop_body)
    # hoisted, not_hoisted = hoist(parfor_params, loop_body,
    #                             typemap, wrapped_blocks)
    setitems = set()
    find_setitems_body(setitems, loop_body, typemap)

    hoisted = []
    not_hoisted = []

    start_block = gufunc_ir.blocks[min(gufunc_ir.blocks.keys())]
    start_block.body = start_block.body[:-1] + hoisted + [start_block.body[-1]]
    unwrap_loop_body(loop_body)

    # store hoisted into diagnostics
    diagnostics = lowerer.metadata["parfor_diagnostics"]
    diagnostics.hoist_info[parfor.id] = {"hoisted": hoisted, "not_hoisted": not_hoisted}

    lowerer.metadata["parfor_diagnostics"].extra_info[str(parfor.id)] = str(
        dpctl.get_current_queue().get_sycl_device().name
    )

    if config.DEBUG_ARRAY_OPT:
        print("After hoisting")
        _print_body(loop_body)

    # Search all the block in the gufunc outline for the sentinel assignment.
    for label, block in gufunc_ir.blocks.items():
        for i, inst in enumerate(block.body):
            if isinstance(inst, ir.Assign) and inst.target.name == sentinel_name:
                # We found the sentinel assignment.
                loc = inst.loc
                scope = block.scope
                # split block across __sentinel__
                # A new block is allocated for the statements prior to the
                # sentinel but the new block maintains the current block label.
                prev_block = ir.Block(scope, loc)
                prev_block.body = block.body[:i]

                # The current block is used for statements after the sentinel.
                block.body = block.body[i + 1 :]
                # But the current block gets a new label.
                body_first_label = min(loop_body.keys())

                # The previous block jumps to the minimum labelled block of the
                # parfor body.
                prev_block.append(ir.Jump(body_first_label, loc))
                # Add all the parfor loop body blocks to the gufunc function's
                # IR.
                for (l, b) in loop_body.items():
                    gufunc_ir.blocks[l] = b
                body_last_label = max(loop_body.keys())
                gufunc_ir.blocks[new_label] = block
                gufunc_ir.blocks[label] = prev_block
                # Add a jump from the last parfor body block to the block
                # containing statements after the sentinel.
                gufunc_ir.blocks[body_last_label].append(ir.Jump(new_label, loc))
                break
        else:
            continue
        break

    if config.DEBUG_ARRAY_OPT:
        print("gufunc_ir last dump before renaming")
        gufunc_ir.dump()

    gufunc_ir.blocks = rename_labels(gufunc_ir.blocks)
    remove_dels(gufunc_ir.blocks)

    if config.DEBUG_ARRAY_OPT:
        sys.stdout.flush()

    if config.DEBUG_ARRAY_OPT:
        print("gufunc_ir last dump")
        gufunc_ir.dump()
        print("flags", flags)
        print("typemap", typemap)

    old_alias = flags.noalias
    if not has_aliases:
        if config.DEBUG_ARRAY_OPT:
            print("No aliases found so adding noalias flag.")
        flags.noalias = True

    remove_dead(gufunc_ir.blocks, gufunc_ir.arg_names, gufunc_ir, typemap)

    if config.DEBUG_ARRAY_OPT:
        print("gufunc_ir after remove dead")
        gufunc_ir.dump()

    kernel_sig = signature(types.none, *gufunc_param_types)

    if config.DEBUG_ARRAY_OPT:
        sys.stdout.flush()

    if config.DEBUG_ARRAY_OPT:
        print("before DUFunc inlining".center(80, "-"))
        gufunc_ir.dump()

    # Inlining all DUFuncs
    dufunc_inliner(
        gufunc_ir,
        lowerer.fndesc.calltypes,
        typemap,
        lowerer.context.typing_context,
        lowerer.context,
    )

    if config.DEBUG_ARRAY_OPT:
        print("after DUFunc inline".center(80, "-"))
        gufunc_ir.dump()

    kernel_func = numba_dppy.compiler.compile_kernel_parfor(
        dpctl.get_current_queue(),
        gufunc_ir,
        gufunc_param_types,
        param_types_addrspaces,
        debug=flags.debuginfo,
    )

    flags.noalias = old_alias

    if config.DEBUG_ARRAY_OPT:
        print("kernel_sig = ", kernel_sig)

    return kernel_func, parfor_args, kernel_sig, func_arg_types, setitems


def _lower_parfor_gufunc(lowerer, parfor):
    """Lowerer that handles LLVM code generation for parfor.
    This function lowers a parfor IR node to LLVM.
    The general approach is as follows:
    1) The code from the parfor's init block is lowered normally
       in the context of the current function.
    2) The body of the parfor is transformed into a gufunc function.
    3) Code is inserted into the main function that calls do_scheduling
       to divide the iteration space for each thread, allocates
       reduction arrays, calls the gufunc function, and then invokes
       the reduction function across the reduction arrays to produce
       the final reduction values.
    """

    typingctx = lowerer.context.typing_context
    targetctx = lowerer.context
    # We copy the typemap here because for race condition variable we'll
    # update their type to array so they can be updated by the gufunc.
    orig_typemap = lowerer.fndesc.typemap
    # replace original typemap with copy and restore the original at the end.
    lowerer.fndesc.typemap = copy.copy(orig_typemap)
    if config.DEBUG_ARRAY_OPT:
        print("lowerer.fndesc", lowerer.fndesc, type(lowerer.fndesc))
    typemap = lowerer.fndesc.typemap
    varmap = lowerer.varmap

    if config.DEBUG_ARRAY_OPT:
        print("_lower_parfor_parallel")
        parfor.dump()

    loc = parfor.init_block.loc
    scope = parfor.init_block.scope

    # produce instructions for init_block
    if config.DEBUG_ARRAY_OPT:
        print("init_block = ", parfor.init_block, type(parfor.init_block))
    for instr in parfor.init_block.body:
        if config.DEBUG_ARRAY_OPT:
            print("lower init_block instr = ", instr)
        lowerer.lower_inst(instr)

    for racevar in parfor.races:
        if racevar not in varmap:
            rvtyp = typemap[racevar]
            rv = ir.Var(scope, racevar, loc)
            lowerer._alloca_var(rv.name, rvtyp)

    alias_map = {}
    arg_aliases = {}
    numba.parfors.parfor.find_potential_aliases_parfor(
        parfor, parfor.params, typemap, lowerer.func_ir, alias_map, arg_aliases
    )
    if config.DEBUG_ARRAY_OPT:
        print("alias_map", alias_map)
        print("arg_aliases", arg_aliases)

    # run get_parfor_outputs() and get_parfor_reductions() before
    # gufunc creation since Jumps are modified so CFG of loop_body
    # dict will become invalid
    assert parfor.params is not None

    parfor_output_arrays = numba.parfors.parfor.get_parfor_outputs(
        parfor, parfor.params
    )

    # compile parfor body as a separate function to be used with GUFuncWrapper
    flags = copy.copy(parfor.flags)
    flags.error_model = "numpy"

    # Can't get here unless flags.set('auto_parallel', ParallelOptions(True))
    index_var_typ = typemap[parfor.loop_nests[0].index_variable.name]

    # index variables should have the same type, check rest of indices
    for l in parfor.loop_nests[1:]:
        assert typemap[l.index_variable.name] == index_var_typ

    numba.parfors.parfor.sequential_parfor_lowering = True
    loop_ranges = [(l.start, l.stop, l.step) for l in parfor.loop_nests]

    try:
        (
            func,
            func_args,
            func_sig,
            func_arg_types,
            modified_arrays,
        ) = _create_gufunc_for_parfor_body(
            lowerer,
            parfor,
            typemap,
            typingctx,
            targetctx,
            flags,
            loop_ranges,
            {},
            bool(alias_map),
            index_var_typ,
            parfor.races,
        )
    finally:
        numba.parfors.parfor.sequential_parfor_lowering = False

    # get the shape signature
    get_shape_classes = parfor.get_shape_classes

    num_inputs = len(func_args) - len(parfor_output_arrays)
    if config.DEBUG_ARRAY_OPT:
        print("func", func, type(func))
        print("func_args", func_args, type(func_args))
        print("func_sig", func_sig, type(func_sig))
        print("num_inputs = ", num_inputs)
        print("parfor_outputs = ", parfor_output_arrays)

    # call the func in parallel by wrapping it with ParallelGUFuncBuilder
    if config.DEBUG_ARRAY_OPT:
        print("loop_nests = ", parfor.loop_nests)
        print("loop_ranges = ", loop_ranges)

    gu_signature = _create_shape_signature(
        parfor.get_shape_classes, num_inputs, func_args, func_sig, parfor.races, typemap
    )

    generate_kernel_launch_ops(
        lowerer,
        func,
        gu_signature,
        func_sig,
        func_args,
        num_inputs,
        func_arg_types,
        loop_ranges,
        modified_arrays,
    )

    if config.DEBUG_ARRAY_OPT:
        sys.stdout.flush()

    # Restore the original typemap of the function that was replaced
    # temporarily at the beginning of this function.
    lowerer.fndesc.typemap = orig_typemap


def _create_shape_signature(
    get_shape_classes,
    num_inputs,
    # num_reductions,
    args,
    func_sig,
    races,
    typemap,
):
    """Create shape signature for GUFunc"""
    if config.DEBUG_ARRAY_OPT:
        print("_create_shape_signature", num_inputs, args)
        arg_start_print = 0
        for i in args[arg_start_print:]:
            print("argument", i, type(i), get_shape_classes(i, typemap=typemap))

    # num_inouts = len(args) - num_reductions
    num_inouts = len(args)
    # maximum class number for array shapes
    classes = [
        get_shape_classes(var, typemap=typemap) if var not in races else (-1,)
        for var in args[1:]
    ]
    class_set = set()
    for _class in classes:
        if _class:
            for i in _class:
                class_set.add(i)
    max_class = max(class_set) + 1 if class_set else 0
    classes.insert(0, (max_class,))  # force set the class of 'sched' argument
    class_set.add(max_class)
    class_map = {}
    # TODO: use prefix + class number instead of single char
    alphabet = ord("a")
    for n in class_set:
        if n >= 0:
            class_map[n] = chr(alphabet)
            alphabet += 1

    alpha_dict = {"latest_alpha": alphabet}

    def bump_alpha(c, class_map):
        if c >= 0:
            return class_map[c]
        else:
            alpha_dict["latest_alpha"] += 1
            return chr(alpha_dict["latest_alpha"])

    gu_sin = []
    gu_sout = []
    count = 0
    syms_sin = ()

    if config.DEBUG_ARRAY_OPT:
        print("args", args)
        print("classes", classes)

    for cls, arg in zip(classes, args):
        count = count + 1
        if cls:
            dim_syms = tuple(bump_alpha(c, class_map) for c in cls)
        else:
            dim_syms = ()
        gu_sin.append(dim_syms)
        syms_sin += dim_syms
    return (gu_sin, gu_sout)


# Keep all the dppy kernels and programs created alive indefinitely.
keep_alive_kernels = []


def generate_kernel_launch_ops(
    lowerer,
    cres,
    gu_signature,
    outer_sig,
    expr_args,
    num_inputs,
    expr_arg_types,
    loop_ranges,
    modified_arrays,
):
    """
    Adds the call to the gufunc function from the main function.
    """
    context = lowerer.context
    builder = lowerer.builder
    sin, sout = gu_signature
    num_dim = len(loop_ranges)

    if config.DEBUG_ARRAY_OPT:
        print("generate_kernel_launch_ops")
        print("args = ", expr_args)
        print(
            "outer_sig = ",
            outer_sig.args,
            outer_sig.return_type,
            outer_sig.recvr,
            outer_sig.pysig,
        )
        print("loop_ranges = ", loop_ranges)
        print("expr_args", expr_args)
        print("expr_arg_types", expr_arg_types)
        print("gu_signature", gu_signature)
        print("sin", sin)
        print("sout", sout)
        print("cres", cres, type(cres))
        print("modified_arrays", modified_arrays)

    # get dppy_cpu_portion_lowerer object
    kernel_launcher = KernelLaunchOps(lowerer, cres.kernel, num_inputs)

    # Get a pointer to the current queue
    curr_queue = kernel_launcher.get_current_queue()

    # Compute number of args
    num_expanded_args = 0

    for arg_type in expr_arg_types:
        if isinstance(arg_type, types.npytypes.Array):
            num_expanded_args += 5 + (2 * arg_type.ndim)
        else:
            num_expanded_args += 1

    if config.DEBUG_ARRAY_OPT:
        print("num_expanded_args = ", num_expanded_args)

    # now that we know the total number of kernel args, lets allocate
    # a kernel_arg array
    kernel_launcher.allocate_kernel_arg_array(num_expanded_args)

    ninouts = len(expr_args)

    def getvar_or_none(lowerer, x):
        try:
            return lowerer.getvar(x)
        except:
            return None

    def loadvar_or_none(lowerer, x):
        try:
            return lowerer.loadvar(x)
        except:
            return None

    def val_type_or_none(context, lowerer, x):
        try:
            return context.get_value_type(lowerer.fndesc.typemap[x])
        except:
            return None

    all_llvm_args = [getvar_or_none(lowerer, x) for x in expr_args[:ninouts]]
    all_val_types = [val_type_or_none(context, lowerer, x) for x in expr_args[:ninouts]]
    all_args = [loadvar_or_none(lowerer, x) for x in expr_args[:ninouts]]

    keep_alive_kernels.append(cres)

    # Call clSetKernelArg for each arg and create arg array for
    # the enqueue function. Put each part of each argument into
    # kernel_arg_array.
    for var, llvm_arg, arg_type, gu_sig, val_type, index in zip(
        expr_args,
        all_llvm_args,
        expr_arg_types,
        sin + sout,
        all_val_types,
        range(len(expr_args)),
    ):

        if config.DEBUG_ARRAY_OPT:
            print(
                "var:",
                var,
                type(var),
                "\n\tllvm_arg:",
                llvm_arg,
                type(llvm_arg),
                "\n\targ_type:",
                arg_type,
                type(arg_type),
                "\n\tgu_sig:",
                gu_sig,
                "\n\tval_type:",
                val_type,
                type(val_type),
                "\n\tindex:",
                index,
            )
        kernel_launcher.process_kernel_arg(
            var, llvm_arg, arg_type, index, modified_arrays, curr_queue
        )

    # loadvars for loop_ranges
    def load_range(v):
        if isinstance(v, ir.Var):
            return lowerer.loadvar(v.name)
        else:
            return context.get_constant(types.uintp, v)

    num_dim = len(loop_ranges)
    for i in range(num_dim):
        start, stop, step = loop_ranges[i]
        start = load_range(start)
        stop = load_range(stop)
        assert step == 1  # We do not support loop steps other than 1
        step = load_range(step)
        loop_ranges[i] = (start, stop, step)

    kernel_launcher.enqueue_kernel_and_copy_back(loop_ranges, curr_queue)

    # At this point we can free the DPCTLSyclQueueRef (curr_queue)
    kernel_launcher.free_queue(sycl_queue_val=curr_queue)


from numba.core.lowering import Lower


class CopyIRException(RuntimeError):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


def relatively_deep_copy(obj, memo):
    # WARNING: there are some issues with genarators which were not investigated
    # and root cause is not found. Though copied IR seems to work fine there are
    # some extra references kept on generator objects which may result in a
    # memory leak.

    obj_id = id(obj)
    if obj_id in memo:
        return memo[obj_id]

    from numba.core.dispatcher import _DispatcherBase
    from numba.core.types.functions import Function, Dispatcher
    from numba.core.bytecode import FunctionIdentity
    from numba.core.typing.templates import Signature
    from numba_dppy.compiler import DPPYFunctionTemplate
    from numba.core.compiler import CompileResult
    from numba.np.ufunc.dufunc import DUFunc
    from ctypes import _CFuncPtr
    from types import ModuleType
    from numba.core.types.abstract import Type

    # objects which shouldn't or can't be copied and it's ok not to copy it.
    if isinstance(
        obj,
        (
            FunctionIdentity,
            _DispatcherBase,
            Function,
            Type,
            Dispatcher,
            ModuleType,
            Signature,
            DPPYFunctionTemplate,
            CompileResult,
            DUFunc,
            _CFuncPtr,
            type,
            str,
            bool,
            type(None),
        ),
    ):
        return obj

    from numba.core.ir import Global, FreeVar
    from numba.core.ir import FunctionIR
    from numba.core.postproc import PostProcessor
    from numba.core.funcdesc import FunctionDescriptor

    if isinstance(obj, FunctionDescriptor):
        cpy = FunctionDescriptor(
            native=obj.native,
            modname=obj.modname,
            qualname=obj.qualname,
            unique_name=obj.unique_name,
            doc=obj.doc,
            typemap=relatively_deep_copy(obj.typemap, memo),
            restype=obj.restype,
            calltypes=relatively_deep_copy(obj.calltypes, memo),
            args=obj.args,
            kws=obj.kws,
            mangler=None,
            argtypes=relatively_deep_copy(obj.argtypes, memo),
            inline=obj.inline,
            noalias=obj.noalias,
            env_name=obj.env_name,
            global_dict=obj.global_dict,
        )
        # mangler parameter is not saved in FunctionDescriptor, but used to generated name.
        # So pass None as mangler parameter and then copy mangled_name by hands
        cpy.mangled_name = obj.mangled_name

        memo[obj_id] = cpy

        return cpy

    if isinstance(obj, FunctionIR):
        # PostProcessor do the following:
        # 1. canonicolize cfg, modifying IR
        # 2. fills internal generators status
        # 3. creates and fills VariableLifetime object
        # We can't copy this objects. So in order to have copy of it we need run PostProcessor on copied IR.
        # This means, that in case PostProcess wasn't run for original object copied object would defer.
        # In order to avoid this we are running PostProcess on original object firstly.
        # This means that copy of IR actually has a side effect on it.
        pp = PostProcessor(obj)
        pp.run()
        cpy = FunctionIR(
            blocks=relatively_deep_copy(obj.blocks, memo),
            is_generator=relatively_deep_copy(obj.is_generator, memo),
            func_id=relatively_deep_copy(obj.func_id, memo),
            loc=obj.loc,
            definitions=relatively_deep_copy(obj._definitions, memo),
            arg_count=obj.arg_count,
            arg_names=relatively_deep_copy(obj.arg_names, memo),
        )
        pp = PostProcessor(cpy)
        pp.run()

        memo[obj_id] = cpy

        return cpy

    if isinstance(obj, Global):
        cpy = Global(name=obj.name, value=obj.value, loc=obj.loc)
        memo[obj_id] = cpy

        return cpy

    if isinstance(obj, FreeVar):
        cpy = FreeVar(index=obj.index, name=obj.name, value=obj.value, loc=obj.loc)
        memo[obj_id] = cpy

        return cpy

    # for containers we need to copy container itself first. And then fill it with copied items.
    if isinstance(obj, list):
        cpy = copy.copy(obj)
        cpy.clear()
        for item in obj:
            cpy.append(relatively_deep_copy(item, memo))
        memo[obj_id] = cpy
        return cpy
    elif isinstance(obj, dict):
        cpy = copy.copy(obj)
        cpy.clear()
        for key, item in obj.items():
            cpy[relatively_deep_copy(key, memo)] = relatively_deep_copy(item, memo)
        memo[obj_id] = cpy
        return cpy
    elif isinstance(obj, tuple):
        # subclass constructors could have different parameters than superclass.
        # e.g. tuple and namedtuple constructors accepts quite different parameters.
        # it is better to have separate section for namedtuple
        tpl = tuple([relatively_deep_copy(item, memo) for item in obj])
        if type(obj) == tuple:
            cpy = tpl
        else:
            cpy = type(obj)(*tpl)
        memo[obj_id] = cpy
        return cpy
    elif isinstance(obj, set):
        cpy = copy.copy(obj)
        cpy.clear()
        for item in obj:
            cpy.add(relatively_deep_copy(item, memo))
        memo[obj_id] = cpy
        return cpy

    # some python objects are not copyable. In such case exception would be raised
    # it is just a convinient point to find such objects
    try:
        cpy = copy.copy(obj)
    except Exception as e:
        raise e

    # __slots__ for subclass specify only members declared in subclass. So to get all members we need to go through
    # all supeclasses
    def get_slots_members(obj):
        keys = []
        typ = obj
        if not isinstance(typ, type):
            typ = type(obj)

        try:
            if len(typ.__slots__):
                keys.extend(typ.__slots__)
            if len(typ.__bases__):
                for base in typ.__bases__:
                    keys.extend(get_slots_members(base))
        except:
            pass

        return keys

    memo[obj_id] = cpy
    keys = []

    # Objects have either __dict__ or __slots__ or neither.
    # If object has none of it and it is copyable we already made a copy, just return it
    # If object is not copyable we shouldn't reach this point.
    try:
        keys = obj.__dict__.keys()
    except:
        try:
            obj.__slots__
            keys = get_slots_members(obj)
        except:
            return cpy

    for key in keys:
        attr = getattr(obj, key)
        attr_cpy = relatively_deep_copy(attr, memo)
        setattr(cpy, key, attr_cpy)

    return cpy


class WrapperDefaultLower(Lower):
    @property
    def _disable_sroa_like_opt(self):
        """For numba_dppy's case we always return True."""
        return True


class DPPYLower(Lower):
    def __init__(self, context, library, fndesc, func_ir, metadata=None):
        Lower.__init__(self, context, library, fndesc, func_ir, metadata)
        memo = {}

        fndesc_cpu = relatively_deep_copy(fndesc, memo)
        func_ir_cpu = relatively_deep_copy(func_ir, memo)

        cpu_context = (
            context.cpu_context if isinstance(context, DPPYTargetContext) else context
        )
        self.gpu_lower = self._lower(context, library, fndesc, func_ir, metadata)
        self.cpu_lower = self._lower(
            cpu_context, library, fndesc_cpu, func_ir_cpu, metadata
        )

    def _lower(self, context, library, fndesc, func_ir, metadata):
        """Create Lower with changed linkageName in debug info"""
        lower = WrapperDefaultLower(context, library, fndesc, func_ir, metadata)

        # Debuginfo
        if context.enable_debuginfo:
            from numba.core.funcdesc import qualifying_prefix, default_mangler
            from numba_dppy.dppy_debuginfo import DPPYDIBuilder

            qualprefix = qualifying_prefix(fndesc.modname, fndesc.qualname)
            mangled_qualname = default_mangler(qualprefix, fndesc.argtypes)

            lower.debuginfo = DPPYDIBuilder(
                module=lower.module,
                filepath=func_ir.loc.filename,
                linkage_name=mangled_qualname,
            )

        return lower

    def lower(self):
        """Numba-dppy's custom lowering function.

        The lowerer has a builtin fallback mechanism for parfor functions.
        We first try to lower a parfor onto a SYCL device using numba-dppy's
        pipeline, if the lowering fails then we fallback to the default Numba
        lowering to CPU. The lowering follow the following steps:

        1. Start lowering of parent function
        2. Try to lower parfor onto the specified SYCL device
            2.a. The ``lower_parfor_rollback`` function prepares function to
                 lower onto to the specified SYCL device and inserts the
                 ``get_global_id`` intrinsic function.
                2.a.a. Start lowering the parfor body and execute
                       ``DPPYLower.lower()`` again.
                2.a.b. If the lowering fails, throw an exception.
            2.b. The ``lower_parfor_rollback`` catches the exception and
                 restores the parfor body to its initial state.
            2.c. Then throw an exception inside ``lower_parfor_rollback``
                 that will be caught inside ``DPPYLower.lower()``.
        3. Catch exception and start parfor lowering with the default Numba CPU
           context.

        TODO/FIXME The rollback approach only works in case no device specific
        modifications were added to function containing the parfor node. If the
        function has any device specific modifications, a different solution
        should be used.

        Raises:
            Exception: If a parfor node could not be lowered to a SYCL device.
        """
        try:
            context = self.gpu_lower.context
            try:
                # Only Numba's CPUContext has the `lower_extension` attribute
                lower_extension_parfor = context.lower_extensions[parfor.Parfor]
                context.lower_extensions[parfor.Parfor] = lower_parfor_rollback
            except Exception as e:
                if config.DEBUG:
                    print(e)
                pass

            self.gpu_lower.lower()
            # if lower does not crash, and parfor_diagnostics is empty then it
            # is a kernel function.
            if not self.gpu_lower.metadata["parfor_diagnostics"].extra_info:
                str_name = dpctl.get_current_queue().get_sycl_device().filter_string
                self.gpu_lower.metadata["parfor_diagnostics"].extra_info[
                    "kernel"
                ] = str_name
            self.base_lower = self.gpu_lower

            try:
                context.lower_extensions[parfor.Parfor] = lower_extension_parfor
            except Exception as e:
                if config.DEBUG:
                    print(e)
                pass
        except Exception as e:
            if config.DEBUG:
                import traceback

                device_filter_str = (
                    dpctl.get_current_queue().get_sycl_device().filter_string
                )
                print(
                    "Failed to offload parfor to " + device_filter_str + ". Due to:\n",
                    e,
                )
                print(traceback.format_exc())

            if config.FALLBACK_ON_CPU == 1:
                self.cpu_lower.context.lower_extensions[
                    parfor.Parfor
                ] = _lower_parfor_parallel
                self.cpu_lower.lower()
                self.base_lower = self.cpu_lower
            else:
                raise e

        self.env = self.base_lower.env
        self.call_helper = self.base_lower.call_helper

    def create_cpython_wrapper(self, release_gil=False):
        return self.base_lower.create_cpython_wrapper(release_gil)


def copy_block(block):
    memo = {}
    new_block = ir.Block(block.scope, block.loc)
    new_block.body = [relatively_deep_copy(stmt, memo) for stmt in block.body]
    return new_block


def lower_parfor_rollback(lowerer, parfor):
    try:
        _lower_parfor_gufunc(lowerer, parfor)
        if config.DEBUG:

            device_filter_str = (
                dpctl.get_current_queue().get_sycl_device().filter_string
            )

            msg = "Parfor offloaded to " + device_filter_str
            print(msg, parfor.loc)
    except Exception as e:

        device_filter_str = dpctl.get_current_queue().get_sycl_device().filter_string
        msg = (
            "Failed to offload parfor to " + device_filter_str + ". Falling "
            "back to default CPU parallelization. Please file a bug report "
            "at https://github.com/IntelPython/numba-dppy. To help us debug "
            "the issue, please add the traceback to the bug report."
        )
        if not config.DEBUG:
            msg += " Set the environment variable NUMBA_DPPY_DEBUG to 1 to "
            msg += "generate a traceback."

        warnings.warn(NumbaPerformanceWarning(msg, parfor.loc))
        raise e


def dppy_lower_array_expr(lowerer, expr):
    raise NotImplementedError(expr)



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import dpctl
import numpy as np
from numba import njit

import numba_dppy as dppy
from numba_dppy.tests._helper import (
    assert_auto_offloading,
    dpnp_debug,
    ensure_dpnp,
    has_gpu_queues,
)


@unittest.skipUnless(
    ensure_dpnp() and has_gpu_queues(),
    "test only when dpnp and GPU is available",
)
class Testdpnp_functions(unittest.TestCase):
    N = 10

    a = np.array(np.random.random(N), dtype=np.float32)
    b = np.array(np.random.random(N), dtype=np.float32)
    tys = [np.int32, np.uint32, np.int64, np.uint64, np.float32, np.double]

    def test_dpnp_interacting_with_parfor(self):
        def f(a, b):
            c = np.sum(a)
            e = np.add(b, a)
            d = c + e
            return d

        device = dpctl.SyclDevice("opencl:gpu")
        with dpctl.device_context(device), assert_auto_offloading(), dpnp_debug():
            njit_f = njit(f)
            got = njit_f(self.a, self.b)
        expected = f(self.a, self.b)

        max_abs_err = got.sum() - expected.sum()
        self.assertTrue(max_abs_err < 1e-4)


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from llvmlite import ir
from numba.core import types
from numba.core.extending import register_jitable
from numba.core.imputils import lower_getattr
from numba.cpython import listobj

ll_void_p = ir.IntType(8).as_pointer()


def get_dpnp_fptr(fn_name, type_names):
    from . import dpnp_fptr_interface as dpnp_glue

    f_ptr = dpnp_glue.get_dpnp_fn_ptr(fn_name, type_names)
    return f_ptr


@register_jitable
def _check_finite_matrix(a):
    for v in np.nditer(a):
        if not np.isfinite(v.item()):
            raise np.linalg.LinAlgError("Array must not contain infs or NaNs.")


@register_jitable
def _dummy_liveness_func(a):
    """pass a list of variables to be preserved through dead code elimination"""
    return a[0]


def dpnp_func(fn_name, type_names, sig):
    f_ptr = get_dpnp_fptr(fn_name, type_names)

    def get_pointer(obj):
        return f_ptr

    return types.ExternalFunctionPointer(sig, get_pointer=get_pointer)


"""
This function retrieves the pointer to the structure where the shape
of an ndarray is stored. We cast it to void * to make it easier to
pass around.
"""


@lower_getattr(types.Array, "shapeptr")
def array_shapeptr(context, builder, typ, value):
    shape_ptr = builder.gep(
        value.operands[0],
        [context.get_constant(types.int32, 0), context.get_constant(types.int32, 5)],
    )

    return builder.bitcast(shape_ptr, ll_void_p)


@lower_getattr(types.List, "size")
def list_size(context, builder, typ, value):
    inst = listobj.ListInstance(context, builder, typ, value)
    return inst.size


@lower_getattr(types.List, "itemsize")
def list_itemsize(context, builder, typ, value):
    llty = context.get_data_type(typ.dtype)
    return context.get_constant(types.uintp, context.get_abi_sizeof(llty))


@lower_getattr(types.List, "ctypes")
def list_ctypes(context, builder, typ, value):
    inst = listobj.ListInstance(context, builder, typ, value)
    return builder.bitcast(inst.data, ll_void_p)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A wrapper to connect to the SPIR-V binaries (Tools, Translator)."""

import os
import shutil
import tempfile
from subprocess import CalledProcessError, check_call

from numba_dppy import config
from numba_dppy.target import LINK_ATOMIC, LLVM_SPIRV_ARGS


def _raise_bad_env_path(msg, path, extra=None):
    error_message = msg.format(path)
    if extra is not None:
        error_message += extra
    raise ValueError(error_message)


_real_check_call = check_call


def check_call(*args, **kwargs):
    return _real_check_call(*args, **kwargs)


class CmdLine:
    def disassemble(self, ipath, opath):
        """
        Disassemble a spirv module.

        Args:
            ipath: Input file path of the spirv module.
            opath: Output file path of the disassembled spirv module.
        """
        flags = []
        check_call(["spirv-dis", *flags, "-o", opath, ipath])

    def validate(self, ipath):
        """
        Validate a spirv module.

        Args:
            ipath: Input file path of the spirv module.
        """
        flags = []
        check_call(["spirv-val", *flags, ipath])

    def optimize(self, ipath, opath):
        """
        Optimize a spirv module.

        Args:
            ipath: Input file path of the spirv module.
            opath: Output file path of the optimized spirv module.
        """
        flags = []
        check_call(["spirv-opt", *flags, "-o", opath, ipath])

    def generate(self, llvm_spirv_args, ipath, opath):
        """
        Generate a spirv module from llvm bitcode.

        We use llvm-spirv tool to translate a llvm bitcode into spirv.

        Args:
            llvm_spirv_args: Args to be provided to llvm-spirv tool.
            ipath: Input file path of the llvm bitcode.
            opath: Output file path of the generated spirv.
        """
        llvm_spirv_flags = []
        if config.DEBUG:
            llvm_spirv_flags.append("--spirv-debug-info-version=ocl-100")

        llvm_spirv_tool = self._llvm_spirv()

        if config.DEBUG:
            print(f"Use llvm-spirv: {llvm_spirv_tool}")

        check_call([llvm_spirv_tool, *llvm_spirv_args, "-o", opath, ipath])

    @staticmethod
    def _llvm_spirv():
        """Return path to llvm-spirv executable."""
        result = None

        if config.LLVM_SPIRV_ROOT:
            result = shutil.which("llvm-spirv", path=config.LLVM_SPIRV_ROOT)

        if not result:
            # use llvm-spirv from dpcpp package.
            # assume dpcpp from .../bin folder.
            # assume llvm-spirv from .../bin-llvm folder.
            bin_llvm = os.path.normpath(
                os.path.dirname(shutil.which("dpcpp")) + "/../bin-llvm/"
            )
            result = shutil.which("llvm-spirv", path=bin_llvm)

        if not result:
            result = "llvm-spirv"

        return result

    def link(self, opath, binaries):
        """
        Link spirv modules.

        Args:
            opath: Output file path of the linked final spirv.
            binaries: Spirv modules to be linked.
        """
        flags = ["--allow-partial-linkage"]
        check_call(["spirv-link", *flags, "-o", opath, *binaries])


class Module(object):
    def __init__(self, context, llvmir, llvmbc):
        """
        Setup
        """
        self._tmpdir = tempfile.mkdtemp()
        self._tempfiles = []
        self._cmd = CmdLine()
        self._finalized = False
        self.context = context

        self._llvmir = llvmir
        self._llvmbc = llvmbc

    def __del__(self):
        # Remove all temporary files
        for afile in self._tempfiles:
            os.unlink(afile)
        # Remove directory
        os.rmdir(self._tmpdir)

    def _create_temp_file(self, name, mode="wb"):
        path = self._track_temp_file(name)
        fobj = open(path, mode=mode)
        return fobj, path

    def _track_temp_file(self, name):
        path = os.path.join(self._tmpdir, "{0}-{1}".format(len(self._tempfiles), name))
        self._tempfiles.append(path)
        return path

    def load_llvm(self):
        """
        Load LLVM with "SPIR-V friendly" SPIR 2.0 spec
        """
        # Create temp file to store the input file
        tmp_llvm_ir, llvm_path = self._create_temp_file("llvm-friendly-spir")
        with tmp_llvm_ir:
            tmp_llvm_ir.write(self._llvmbc)

        self._llvmfile = llvm_path

    def finalize(self):
        """
        Finalize module and return the SPIR-V code
        """
        assert not self._finalized, "Module finalized already"

        # Generate SPIR-V from "friendly" LLVM-based SPIR 2.0
        spirv_path = self._track_temp_file("generated-spirv")

        binary_paths = [spirv_path]

        llvm_spirv_args = []
        for key in list(self.context.extra_compile_options.keys()):
            if key == LINK_ATOMIC:
                from .ocl.atomics import get_atomic_spirv_path

                binary_paths.append(get_atomic_spirv_path())
            elif key == LLVM_SPIRV_ARGS:
                llvm_spirv_args = self.context.extra_compile_options[key]
            del self.context.extra_compile_options[key]

        if config.SAVE_IR_FILES != 0:
            # Dump the llvmir and llvmbc in file
            with open("generated_llvm.ir", "w") as f:
                f.write(self._llvmir)
            with open("generated_llvm.bc", "wb") as f:
                f.write(self._llvmbc)

            print("Generated LLVM IR".center(80, "-"))
            print("generated_llvm.ir")
            print("".center(80, "="))
            print("Generated LLVM Bitcode".center(80, "-"))
            print("generated_llvm.bc")
            print("".center(80, "="))

        self._cmd.generate(
            llvm_spirv_args=llvm_spirv_args, ipath=self._llvmfile, opath=spirv_path
        )

        if len(binary_paths) > 1:
            spirv_path = self._track_temp_file("linked-spirv")
            self._cmd.link(spirv_path, binary_paths)

        if config.SAVE_IR_FILES != 0:
            # Dump the llvmir and llvmbc in file
            with open("generated_spirv.spir", "wb") as f1:
                with open(spirv_path, "rb") as f2:
                    spirv_content = f2.read()
                    f1.write(spirv_content)

            print("Generated SPIRV".center(80, "-"))
            print("generated_spirv.spir")
            print("".center(80, "="))

        # Validate the SPIR-V code
        if config.SPIRV_VAL == 1:
            try:
                self._cmd.validate(ipath=spirv_path)
            except CalledProcessError:
                print("SPIR-V Validation failed...")
                pass
            else:
                # Optimize SPIR-V code
                opt_path = self._track_temp_file("optimized-spirv")
                self._cmd.optimize(ipath=spirv_path, opath=opt_path)

                if config.DUMP_ASSEMBLY:
                    # Disassemble optimized SPIR-V code
                    dis_path = self._track_temp_file("disassembled-spirv")
                    self._cmd.disassemble(ipath=opt_path, opath=dis_path)
                    with open(dis_path, "rb") as fin_opt:
                        print("ASSEMBLY".center(80, "-"))
                        print(fin_opt.read())
                        print("".center(80, "="))

        # Read and return final SPIR-V (not optimized!)
        with open(spirv_path, "rb") as fin:
            spirv = fin.read()

        self._finalized = True

        return spirv


def llvm_to_spirv(context, llvmir, llvmbc):
    """
    Generate SPIR-V from LLVM Bitcode.

    Args:
        context: Numba target context.
        llvmir: LLVM IR.
        llvmbc: LLVM Bitcode.

    Returns:
        spirv: SPIR-V binary.
    """
    mod = Module(context, llvmir, llvmbc)
    mod.load_llvm()
    return mod.finalize()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings

from numba.core import config
from packaging import version


class DpctlMinimumVersionRequiredError(Exception):
    """
    A ``DpctlMinimumVersionRequiredError`` indicates that the version of dpctl
    does not satisfy the minimum version requirement.

    """

    pass


# Check for dpctl 0.8.0 or higher in the system.
_dpctl_found = False
try:
    import dpctl

    # Versions of dpctl lower than 0.8.0 are not compatible with current main
    # of numba_dppy.
    if version.parse(dpctl.__version__) < version.parse("0.8.0"):
        raise DpctlMinimumVersionRequiredError

    # For the Numba_dppy extension to work, we should have at least one
    # non-host SYCL device installed.
    # FIXME: In future, we should support just the host device.
    if not dpctl.select_default_device().is_host:
        _dpctl_found = True
    else:
        msg = "dpctl could not find any non-host SYCL device on the system. "
        msg += "A non-host SYCL device is required to use numba_dppy."
        warnings.warn(msg, UserWarning)
except DpctlMinimumVersionRequiredError:
    msg = "numba_dppy is not compatible with " + dpctl.__version__ + "."
    msg += " Install dpctl 0.8.* or higher."
    warnings.warn(msg, UserWarning)
except:
    msg = "Please install dpctl 0.8.* or higher."
    warnings.warn(msg, UserWarning)

# Set this config flag based on if dpctl is found or not. The config flags is
# used elsewhere inside Numba.
dppy_present = _dpctl_found


def _readenv(name, ctor, default):
    """Original version from numba/core/config.py
    class _EnvReloader():
        ...
        def process_environ():
            def _readenv(): ...
    """
    value = os.environ.get(name)
    if value is None:
        return default() if callable(default) else default
    try:
        return ctor(value)
    except Exception:
        import warnings

        warnings.warn(
            "environ %s defined but failed to parse '%s'" % (name, value),
            RuntimeWarning,
        )
        return default


def __getattr__(name):
    """Fallback to Numba config"""
    return getattr(config, name)


# Save intermediate files being generated by DPPY
SAVE_IR_FILES = _readenv("NUMBA_DPPY_SAVE_IR_FILES", int, 0)

# Turn SPIRV-VALIDATION ON/OFF switch
SPIRV_VAL = _readenv("NUMBA_DPPY_SPIRV_VAL", int, 0)

# Dump offload diagnostics
OFFLOAD_DIAGNOSTICS = _readenv("NUMBA_DPPY_OFFLOAD_DIAGNOSTICS", int, 0)

FALLBACK_ON_CPU = _readenv("NUMBA_DPPY_FALLBACK_ON_CPU", int, 1)

# Activate Native floating point atomcis support for supported devices.
# Requires llvm-spirv supporting the FP atomics extension
NATIVE_FP_ATOMICS = _readenv("NUMBA_DPPY_ACTIVATE_ATOMICS_FP_NATIVE", int, 0)
LLVM_SPIRV_ROOT = _readenv("NUMBA_DPPY_LLVM_SPIRV_ROOT", str, "")
# Emit debug info
DEBUG = _readenv("NUMBA_DPPY_DEBUG", int, config.DEBUG)
DEBUGINFO_DEFAULT = _readenv("NUMBA_DPPY_DEBUGINFO", int, config.DEBUGINFO_DEFAULT)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from contextlib import contextmanager

import dpctl
from numba import njit
from numba.core.dispatcher import TargetConfig
from numba.core.retarget import BasicRetarget

from numba_dppy.target import DPPY_TARGET_NAME


class DPPYRetarget(BasicRetarget):
    def __init__(self, filter_str):
        self.filter_str = filter_str
        super(DPPYRetarget, self).__init__()

    @property
    def output_target(self):
        return DPPY_TARGET_NAME

    def compile_retarget(self, cpu_disp):
        kernel = njit(_target=DPPY_TARGET_NAME)(cpu_disp.py_func)
        return kernel


_first_level_cache = dict()


def _retarget(sycl_queue):
    filter_string = sycl_queue.sycl_device.filter_string

    result = _first_level_cache.get(filter_string)

    if not result:
        result = DPPYRetarget(filter_string)
        _first_level_cache[filter_string] = result

    return result


def _retarget_context_manager(sycl_queue):
    """Return context manager for retargeting njit offloading."""
    retarget = _retarget(sycl_queue)
    return TargetConfig.switch_target(retarget)


def _register_context_factory():
    dpctl.nested_context_factories.append(_retarget_context_manager)


_register_context_factory()
offload_to_sycl_device = dpctl.device_context



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
import dpctl.memory as dpctl_mem
import dpctl.tensor as dpt
import numpy as np
import pytest

from numba_dppy.tests._helper import skip_test
from numba_dppy.utils import as_usm_obj, copy_to_numpy_from_usm_obj, has_usm_memory

from . import _helper


def test_has_usm_memory(offload_device):
    if skip_test(offload_device):
        pytest.skip()

    a = np.ones(1023, dtype=np.float32)

    with dpctl.device_context(offload_device):
        # test usm_ndarray
        da = dpt.usm_ndarray(a.shape, dtype=a.dtype, buffer="shared")
        usm_mem = has_usm_memory(da)
        assert da.usm_data._pointer == usm_mem._pointer

        # test usm allocated numpy.ndarray
        buf = dpctl_mem.MemoryUSMShared(a.size * a.dtype.itemsize)
        ary_buf = np.ndarray(a.shape, buffer=buf, dtype=a.dtype)
        usm_mem = has_usm_memory(ary_buf)
        assert buf._pointer == usm_mem._pointer

        usm_mem = has_usm_memory(a)
        assert usm_mem is None


def test_as_usm_obj(offload_device):
    if skip_test(offload_device):
        pytest.skip()

    a = np.ones(1023, dtype=np.float32)
    b = a * 3

    with dpctl.device_context(offload_device) as queue:
        a_copy = np.empty_like(a)
        usm_mem = as_usm_obj(a, queue=queue)
        copy_to_numpy_from_usm_obj(usm_mem, a_copy)
        assert np.all(a == a_copy)

        b_copy = np.empty_like(b)
        usm_mem = as_usm_obj(b, queue=queue, copy=False)
        copy_to_numpy_from_usm_obj(usm_mem, b_copy)
        assert np.any(np.not_equal(b, b_copy))



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
import numpy as np
import pytest

import numba_dppy as dppy
from numba_dppy.tests._helper import ensure_dpnp, skip_test

list_of_dtype = [
    np.int32,
    np.int64,
    np.float32,
    np.float64,
]


@pytest.fixture(params=list_of_dtype)
def dtype(request):
    return request.param


list_of_usm_type = [
    "shared",
    "device",
    "host",
]


@pytest.fixture(params=list_of_usm_type)
def usm_type(request):
    return request.param


def test_dpnp_create_array_in_context(offload_device, dtype):
    if not ensure_dpnp():
        pytest.skip("No DPNP")

    import dpnp

    if skip_test(offload_device):
        pytest.skip("No device for " + offload_device)

    if (
        "opencl" not in dpctl.get_current_queue().sycl_device.filter_string
        and "opencl" in offload_device
    ):
        pytest.skip("Bug in DPNP. See: IntelPython/dpnp#723")

    with dpctl.device_context(offload_device):
        a = dpnp.arange(1024, dtype=dtype)  # noqa


def test_consuming_array_from_dpnp(offload_device, dtype):
    if not ensure_dpnp():
        pytest.skip("No DPNP")

    import dpnp

    if skip_test(offload_device):
        pytest.skip("No device for " + offload_device)

    if (
        "opencl" not in dpctl.get_current_queue().sycl_device.filter_string
        and "opencl" in offload_device
    ):
        pytest.skip("Bug in DPNP. See: IntelPython/dpnp#723")

    @dppy.kernel
    def data_parallel_sum(a, b, c):
        """
        Vector addition using the ``kernel`` decorator.
        """
        i = dppy.get_global_id(0)
        c[i] = a[i] + b[i]

    global_size = 1021

    with dpctl.device_context(offload_device):
        a = dppy.asarray(dpnp.arange(global_size, dtype=dtype))
        b = dppy.asarray(dpnp.arange(global_size, dtype=dtype))
        c = dppy.asarray(dpnp.ones_like(a))

        data_parallel_sum[global_size, dppy.DEFAULT_LOCAL_SIZE](a, b, c)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import warnings

import numpy
from numba.core import types
from numba.core.imputils import Registry
from numba.core.itanium_mangler import mangle

from .oclimpl import _declare_function

registry = Registry()
lower = registry.lower

# -----------------------------------------------------------------------------

_unary_b_f = types.int32(types.float32)
_unary_b_d = types.int32(types.float64)
_unary_f_f = types.float32(types.float32)
_unary_d_d = types.float64(types.float64)
_binary_f_ff = types.float32(types.float32, types.float32)
_binary_d_dd = types.float64(types.float64, types.float64)

_binary_f_fi = types.float32(types.float32, types.int32)
_binary_f_ii = types.float32(types.int32, types.int32)
_binary_f_fl = types.float32(types.float32, types.int64)
_binary_f_fd = types.float32(types.float32, types.float64)
_binary_d_di = types.float64(types.float64, types.int32)
_binary_d_dl = types.float64(types.float64, types.int64)

sig_mapper = {
    "f->f": _unary_f_f,
    "d->d": _unary_d_d,
    "ff->f": _binary_f_ff,
    "dd->d": _binary_d_dd,
    "fi->f": _binary_f_fi,
    "fl->f": _binary_f_fl,
    "ff->f": _binary_f_ff,
    "di->d": _binary_d_di,
    "dl->d": _binary_d_dl,
    "dd->d": _binary_d_dd,
}

function_descriptors = {
    "isnan": (_unary_b_f, _unary_b_d),
    "isinf": (_unary_b_f, _unary_b_d),
    "ceil": (_unary_f_f, _unary_d_d),
    "floor": (_unary_f_f, _unary_d_d),
    "trunc": (_unary_f_f, _unary_d_d),
    "fabs": (_unary_f_f, _unary_d_d),
    "sqrt": (_unary_f_f, _unary_d_d),
    "exp": (_unary_f_f, _unary_d_d),
    "expm1": (_unary_f_f, _unary_d_d),
    "log": (_unary_f_f, _unary_d_d),
    "log10": (_unary_f_f, _unary_d_d),
    "log1p": (_unary_f_f, _unary_d_d),
    "sin": (_unary_f_f, _unary_d_d),
    "cos": (_unary_f_f, _unary_d_d),
    "tan": (_unary_f_f, _unary_d_d),
    "asin": (_unary_f_f, _unary_d_d),
    "acos": (_unary_f_f, _unary_d_d),
    "atan": (_unary_f_f, _unary_d_d),
    "sinh": (_unary_f_f, _unary_d_d),
    "cosh": (_unary_f_f, _unary_d_d),
    "tanh": (_unary_f_f, _unary_d_d),
    "asinh": (_unary_f_f, _unary_d_d),
    "acosh": (_unary_f_f, _unary_d_d),
    "atanh": (_unary_f_f, _unary_d_d),
    "copysign": (_binary_f_ff, _binary_d_dd),
    "atan2": (_binary_f_ff, _binary_d_dd),
    "pow": (_binary_f_ff, _binary_d_dd),
    "fmod": (_binary_f_ff, _binary_d_dd),
    "erf": (_unary_f_f, _unary_d_d),
    "erfc": (_unary_f_f, _unary_d_d),
    "gamma": (_unary_f_f, _unary_d_d),
    "lgamma": (_unary_f_f, _unary_d_d),
    "ldexp": (_binary_f_fi, _binary_f_fl, _binary_d_di, _binary_d_dl),
    "hypot": (_binary_f_fi, _binary_f_ff, _binary_d_dl, _binary_d_dd),
    "exp2": (_unary_f_f, _unary_d_d),
    "log2": (_unary_f_f, _unary_d_d),
    # unsupported functions listed in the math module documentation:
    # frexp, ldexp, trunc, modf, factorial, fsum
}


# some functions may be named differently by the underlying math
# library as oposed to the Python name.
_lib_counterpart = {"gamma": "tgamma"}


def _mk_fn_decl(name, decl_sig):
    sym = _lib_counterpart.get(name, name)

    def core(context, builder, sig, args):
        fn = _declare_function(
            context, builder, sym, decl_sig, decl_sig.args, mangler=mangle
        )
        res = builder.call(fn, args)
        return context.cast(builder, res, decl_sig.return_type, sig.return_type)

    core.__name__ = name
    return core


_supported = [
    "sin",
    "cos",
    "tan",
    "asin",
    "acos",
    "atan",
    "atan2",
    "sinh",
    "cosh",
    "tanh",
    "asinh",
    "acosh",
    "atanh",
    "isnan",
    "isinf",
    "ceil",
    "floor",
    "fabs",
    "sqrt",
    "exp",
    "expm1",
    "log",
    "log10",
    "log1p",
    "copysign",
    "pow",
    "fmod",
    "erf",
    "erfc",
    "gamma",
    "lgamma",
    "ldexp",
    "trunc",
    "hypot",
    "exp2",
    "log2",
]


lower_ocl_impl = dict()


def function_name_to_supported_decl(name, sig):
    try:
        # only symbols present in the math module
        key = getattr(math, name)
    except AttributeError:
        try:
            key = getattr(numpy, name)
        except:
            return None

    fn = _mk_fn_decl(name, sig)
    # lower(key, *sig.args)(fn)
    lower_ocl_impl[(name, sig)] = lower(key, *sig.args)(fn)


for name in _supported:
    sigs = function_descriptors.get(name)
    if sigs is None:
        warnings.warn("OCL - failed to register '{0}'".format(name))
        continue

    for sig in sigs:
        function_name_to_supported_decl(name, sig)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
import numpy as np
import pytest

import numba_dppy as dppy
from numba_dppy.tests._helper import skip_test


def call_kernel(global_size, local_size, A, B, C, func):
    func[global_size, local_size](A, B, C)


global_size = 10
local_size = 1
N = global_size * local_size


def sum_kernel(a, b, c):
    i = dppy.get_global_id(0)
    c[i] = a[i] + b[i]


list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


list_of_dtypes = [
    np.float32,
    np.float64,
]


@pytest.fixture(params=list_of_dtypes)
def input_arrays(request):
    # The size of input and out arrays to be used
    a = np.array(np.random.random(N), request.param)
    b = np.array(np.random.random(N), request.param)
    c = np.zeros_like(a)
    return a, b, c


list_of_kernel_opt = [
    {"read_only": ["a", "b"], "write_only": ["c"], "read_write": []},
    {},
]


@pytest.fixture(params=list_of_kernel_opt)
def kernel(request):
    return dppy.kernel(access_types=request.param)(sum_kernel)


def test_kernel_arg_accessor(filter_str, input_arrays, kernel):
    if skip_test(filter_str):
        pytest.skip()

    a, b, actual = input_arrays
    expected = a + b
    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device):
        call_kernel(global_size, local_size, a, b, actual, kernel)
    np.testing.assert_allclose(actual, expected, rtol=1e-5, atol=0)



#! /usr/bin/env python
# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import dpctl
import numpy as np

import numba_dppy as dppy

from . import _helper


@dppy.kernel
def data_parallel_sum(a, b, c):
    i = dppy.get_global_id(0)
    c[i] = a[i] + b[i]


global_size = 64
N = global_size

a = np.array(np.random.random(N), dtype=np.float32)
b = np.array(np.random.random(N), dtype=np.float32)
d = a + b


@unittest.skipUnless(_helper.has_cpu_queues(), "test only on CPU system")
class TestDPPYDeviceArrayArgsGPU(unittest.TestCase):
    def test_device_array_args_cpu(self):
        c = np.ones_like(a)

        with dpctl.device_context("opencl:cpu"):
            data_parallel_sum[global_size, dppy.DEFAULT_LOCAL_SIZE](a, b, c)

            self.assertTrue(np.all(c == d))


@unittest.skipUnless(_helper.has_gpu_queues(), "test only on GPU system")
class TestDPPYDeviceArrayArgsCPU(unittest.TestCase):
    def test_device_array_args_gpu(self):
        c = np.ones_like(a)

        with dpctl.device_context("opencl:gpu"):
            data_parallel_sum[global_size, dppy.DEFAULT_LOCAL_SIZE](a, b, c)

        self.assertTrue(np.all(c == d))


if __name__ == "__main__":
    unittest.main()



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl
import numpy as np
import pytest

import numba_dppy as dppy
from numba_dppy.tests._helper import skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


@pytest.mark.xfail
def test_print_only_str(filter_str):
    try:
        device = dpctl.SyclDevice(filter_str)
        with dpctl.device_context(device):
            pass
    except Exception:
        pytest.skip()

    @dppy.kernel
    def f():
        print("test", "test2")

    # This test will fail, we currently can not print only string.
    # The LLVM generated for printf() function with only string gets
    # replaced by a puts() which fails due to lack of addrspace in the
    # puts function signature right now, and would fail in general due
    # to lack of support for puts() in OpenCL.

    with dpctl.device_context(filter_str):
        f[3, dppy.DEFAULT_LOCAL_SIZE]()


list_of_dtypes = [
    np.int32,
    np.int64,
    np.float32,
    np.float64,
]


@pytest.fixture(params=list_of_dtypes)
def input_arrays(request):
    a = np.array([0], request.param)
    a[0] = 6
    return a


def test_print(filter_str, input_arrays, capfd):
    if skip_test(filter_str):
        pytest.skip()

    @dppy.kernel
    def f(a):
        print("test", a[0])

    a = input_arrays
    global_size = 3

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device):
        f[global_size, dppy.DEFAULT_LOCAL_SIZE](a)
        captured = capfd.readouterr()
        assert "test" in captured.out



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import platform

import dpctl
import numpy as np
import pytest

import numba_dppy as dppy
from numba_dppy.tests._helper import skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


def skip_if_win():
    return platform.system == "Windows"


def test_proper_lowering(filter_str):
    if skip_test(filter_str):
        pytest.skip()

    # We perform eager compilation at the site of
    # @dppy.kernel. This takes the default dpctl
    # queue which is level_zero backed. Level_zero
    # is not yet supported on Windows platform and
    # hence we skip these tests if the platform is
    # Windows regardless of which backend filter_str
    # specifies.
    if skip_if_win():
        pytest.skip()

    # This will trigger eager compilation
    @dppy.kernel("void(float32[::1])")
    def twice(A):
        i = dppy.get_global_id(0)
        d = A[i]
        dppy.barrier(dppy.CLK_LOCAL_MEM_FENCE)  # local mem fence
        A[i] = d * 2

    N = 256
    arr = np.random.random(N).astype(np.float32)
    orig = arr.copy()

    with dpctl.device_context(filter_str):
        twice[N, N // 2](arr)

    # The computation is correct?
    np.testing.assert_allclose(orig * 2, arr)


def test_no_arg_barrier_support(filter_str):
    if skip_test(filter_str):
        pytest.skip()

    if skip_if_win():
        pytest.skip()

    @dppy.kernel("void(float32[::1])")
    def twice(A):
        i = dppy.get_global_id(0)
        d = A[i]
        # no argument defaults to global mem fence
        dppy.barrier()
        A[i] = d * 2

    N = 256
    arr = np.random.random(N).astype(np.float32)
    orig = arr.copy()

    with dpctl.device_context(filter_str):
        twice[N, dppy.DEFAULT_LOCAL_SIZE](arr)

    # The computation is correct?
    np.testing.assert_allclose(orig * 2, arr)


def test_local_memory(filter_str):
    if skip_test(filter_str):
        pytest.skip()
    blocksize = 10

    if skip_if_win():
        pytest.skip()

    @dppy.kernel("void(float32[::1])")
    def reverse_array(A):
        lm = dppy.local.array(shape=10, dtype=np.float32)
        i = dppy.get_global_id(0)

        # preload
        lm[i] = A[i]
        # barrier local or global will both work as we only have one work group
        dppy.barrier(dppy.CLK_LOCAL_MEM_FENCE)  # local mem fence
        # write
        A[i] += lm[blocksize - 1 - i]

    arr = np.arange(blocksize).astype(np.float32)
    orig = arr.copy()

    with dpctl.device_context(filter_str):
        reverse_array[blocksize, blocksize](arr)

    expected = orig[::-1] + orig
    np.testing.assert_allclose(expected, arr)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba import types
from numba.core.typing.templates import AttributeTemplate, infer_getattr

import numba_dppy


@infer_getattr
class DppyDpnpTemplate(AttributeTemplate):
    key = types.Module(numba_dppy)

    def resolve_dpnp(self, mod):
        return types.Module(numba_dppy.dpnp)


"""
This adds a shapeptr attribute to Numba type representing np.ndarray.
This allows us to get the raw pointer to the structure where the shape
of an ndarray is stored from an overloaded implementation
"""


@infer_getattr
class ArrayAttribute(AttributeTemplate):
    key = types.Array

    def resolve_shapeptr(self, ary):
        return types.voidptr


@infer_getattr
class ListAttribute(AttributeTemplate):
    key = types.List

    def resolve_size(self, ary):
        return types.int64

    def resolve_itemsize(self, ary):
        return types.int64

    def resolve_ctypes(self, ary):
        return types.voidptr



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dpctl.tensor as dpt
import numpy as np
import pytest
from numba.misc.special import typeof

from numba_dppy.driver import USMNdArrayType
from numba_dppy.tests._helper import skip_test

list_of_dtypes = [
    np.int32,
    np.float32,
    np.int64,
    np.float64,
]


@pytest.fixture(params=list_of_dtypes)
def dtype(request):
    return request.param


list_of_usm_type = [
    "shared",
    "device",
    "host",
]


@pytest.fixture(params=list_of_usm_type)
def usm_type(request):
    return request.param


def test_usm_ndarray_type(offload_device, dtype, usm_type):
    if skip_test(offload_device):
        pytest.skip()

    a = np.array(np.random.random(10), dtype)
    da = dpt.usm_ndarray(a.shape, dtype=a.dtype, buffer=usm_type)

    assert isinstance(typeof(da), USMNdArrayType)
    assert da.usm_type == usm_type



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import dpctl
import numpy as np
import pytest
from numba import njit

import numba_dppy as dppy
from numba_dppy.tests._helper import dpnp_debug

from ._helper import args_string, wrapper_function
from .dpnp_skip_test import dpnp_skip_test as skip_test

filter_strings = [
    "level_zero:gpu:0",
    "opencl:gpu:0",
    "opencl:cpu:0",
]

filter_strings_with_skips_for_opencl = [
    "level_zero:gpu:0",
    pytest.param("opencl:gpu:0", marks=pytest.mark.skip(reason="Freeze")),
    pytest.param("opencl:cpu:0", marks=pytest.mark.skip(reason="Segmentation fault")),
    # pytest.param("opencl:cpu:0", marks=pytest.mark.xfail(reason="Segmentation fault")),  # run with --boxed
]


# From https://github.com/IntelPython/dpnp/blob/0.4.0/tests/test_linalg.py#L8
def vvsort(val, vec):
    size = val.size
    for i in range(size):
        imax = i
        for j in range(i + 1, size):
            if np.abs(val[imax]) < np.abs(val[j]):
                imax = j

        temp = val[i]
        val[i] = val[imax]
        val[imax] = temp

        if not (vec is None):
            for k in range(size):
                temp = vec[k, i]
                vec[k, i] = vec[k, imax]
                vec[k, imax] = temp


def get_fn(name, nargs):
    args = args_string(nargs)
    return wrapper_function(args, f"np.{name}({args})", globals())


list_of_dtypes = [
    np.int32,
    np.int64,
    np.float32,
    np.float64,
]


@pytest.fixture(params=list_of_dtypes)
def eig_input(request):
    # The size of input and out arrays to be used
    N = 10
    a = np.arange(N * N, dtype=request.param).reshape((N, N))
    symm_a = (
        np.tril(a)
        + np.tril(a, -1).T
        + +np.diag(np.full((N,), N * N, dtype=request.param))
    )
    return symm_a


@pytest.mark.parametrize("filter_str", filter_strings_with_skips_for_opencl)
def test_eig(filter_str, eig_input, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = eig_input
    fn = get_fn("linalg.eig", 1)
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual_val, actual_vec = f(a)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected_val, expected_vec = fn(a)

        # sort val/vec by abs value
        vvsort(actual_val, actual_vec)
        vvsort(expected_val, expected_vec)

        # NP change sign of vectors
        for i in range(expected_vec.shape[1]):
            if expected_vec[0, i] * actual_vec[0, i] < 0:
                expected_vec[:, i] = -expected_vec[:, i]

        assert np.allclose(actual_val, expected_val)
        assert np.allclose(actual_vec, expected_vec)


@pytest.fixture(params=list_of_dtypes)
def dtype(request):
    return request.param


list_of_dim = [
    (10, 1, 10, 1),
    (10, 1, 10, 2),
    (2, 10, 10, 1),
    (10, 2, 2, 10),
]


@pytest.fixture(params=list_of_dim)
def dot_input(request):
    # The size of input and out arrays to be used
    a1, a2, b1, b2 = request.param
    a = np.array(np.random.random(a1 * a2))
    b = np.array(np.random.random(b1 * b2))
    if a2 != 1:
        a = a.reshape(a1, a2)
    if b2 != 1:
        b = b.reshape(b1, b2)

    return a, b


list_of_dot_name = ["dot", "vdot"]


@pytest.fixture(params=list_of_dot_name)
def dot_name(request):
    return request.param


@pytest.mark.parametrize("filter_str", filter_strings_with_skips_for_opencl)
def test_dot(filter_str, dot_name, dot_input, dtype, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a, b = dot_input

    if dot_name == "vdot":
        if a.size != b.size:
            pytest.skip("vdot only supports same sized arrays")

    a = a.astype(dtype)
    b = b.astype(dtype)
    fn = get_fn(dot_name, 2)
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a, b)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected = fn(a, b)
        assert np.allclose(actual, expected)


@pytest.mark.parametrize("filter_str", filter_strings_with_skips_for_opencl)
def test_matmul(filter_str, dtype, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = np.array(np.random.random(10 * 2), dtype=dtype).reshape(10, 2)
    b = np.array(np.random.random(2 * 10), dtype=dtype).reshape(2, 10)
    fn = get_fn("matmul", 2)
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a, b)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected = fn(a, b)
        assert np.allclose(actual, expected)


@pytest.mark.skip(reason="dpnp does not support it yet")
def test_cholesky(filter_str, dtype, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = np.array([[1, -2], [2, 5]], dtype=dtype)
    fn = get_fn("linalg.cholesky", 1)
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected = fn(a)
        assert np.allclose(actual, expected)


list_of_det_input = [
    [[0, 0], [0, 0]],
    [[1, 2], [1, 2]],
    [[1, 2], [3, 4]],
    [[[1, 2], [3, 4]], [[1, 2], [2, 1]], [[1, 3], [3, 1]]],
    [
        [[[1, 2], [3, 4]], [[1, 2], [2, 1]]],
        [[[1, 3], [3, 1]], [[0, 1], [1, 3]]],
    ],
]


@pytest.fixture(params=list_of_det_input)
def det_input(request):
    return request.param


@pytest.mark.parametrize("filter_str", filter_strings)
def test_det(filter_str, det_input, dtype, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = np.array(det_input, dtype=dtype)
    fn = get_fn("linalg.det", 1)
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected = fn(a)
        assert np.allclose(actual, expected)


@pytest.mark.parametrize("filter_str", filter_strings_with_skips_for_opencl)
def test_multi_dot(filter_str, capfd):
    if skip_test(filter_str):
        pytest.skip()

    def fn(A, B, C, D):
        c = np.linalg.multi_dot([A, B, C, D])
        return c

    A = np.random.random((10000, 100))
    B = np.random.random((100, 1000))
    C = np.random.random((1000, 5))
    D = np.random.random((5, 333))
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(A, B, C, D)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected = fn(A, B, C, D)
        assert np.allclose(actual, expected)


list_of_power = [2, 3, 0]


@pytest.fixture(params=list_of_power)
def power(request):
    return request.param


list_of_matrix_power_input = [
    [[0, 0], [0, 0]],
    [[1, 2], [1, 2]],
    [[1, 2], [3, 4]],
]


@pytest.fixture(params=list_of_matrix_power_input)
def matrix_power_input(request):
    return request.param


@pytest.mark.parametrize("filter_str", filter_strings_with_skips_for_opencl)
def test_matrix_power(filter_str, matrix_power_input, power, dtype, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = np.array(matrix_power_input, dtype=dtype)
    fn = get_fn("linalg.matrix_power", 2)
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(a, power)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected = fn(a, power)
        assert np.allclose(actual, expected)


@pytest.mark.parametrize("filter_str", filter_strings)
@pytest.mark.parametrize(
    "matrix_rank_input",
    [
        pytest.param(
            np.eye(4), marks=pytest.mark.xfail(reason="dpnp does not support it yet")
        ),
        np.ones((4,)),
        pytest.param(
            np.ones((4, 4)),
            marks=pytest.mark.xfail(reason="dpnp does not support it yet"),
        ),
        np.zeros((4,)),
    ],
)
def test_matrix_rank(filter_str, matrix_rank_input, capfd):
    if skip_test(filter_str):
        pytest.skip()

    fn = get_fn("linalg.matrix_rank", 1)
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual = f(matrix_rank_input)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected = fn(matrix_rank_input)
        assert np.allclose(actual, expected)


@pytest.mark.parametrize("filter_str", filter_strings_with_skips_for_opencl)
def test_eigvals(filter_str, eig_input, capfd):
    if skip_test(filter_str):
        pytest.skip()

    a = eig_input
    fn = get_fn("linalg.eigvals", 1)
    f = njit(fn)

    device = dpctl.SyclDevice(filter_str)
    with dpctl.device_context(device), dpnp_debug():
        actual_val = f(a)
        captured = capfd.readouterr()
        assert "dpnp implementation" in captured.out

        expected_val = fn(a)

        # sort val/vec by abs value
        vvsort(actual_val, None)
        vvsort(expected_val, None)

        assert np.allclose(actual_val, expected_val)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
from numba import float32, float64, int32, int64, uint32, uint64
from numba.core import types

import numba_dppy as dppy
import numba_dppy.extended_numba_itanium_mangler as itanium_mangler
from numba_dppy.utils import address_space

list_of_dtypes = [
    (int32, "i"),
    (int64, "x"),
    (uint32, "j"),
    (uint64, "y"),
    (float32, "f"),
    (float64, "d"),
]


@pytest.fixture(params=list_of_dtypes)
def dtypes(request):
    return request.param


list_of_addrspaces = [
    (address_space.PRIVATE, "3AS0"),
    (address_space.GLOBAL, "3AS1"),
    (address_space.LOCAL, "3AS3"),
    (address_space.GENERIC, "3AS4"),
]


@pytest.fixture(params=list_of_addrspaces)
def addrspaces(request):
    return request.param


def test_mangling_arg_type(dtypes):
    dtype, expected_str = dtypes
    got = itanium_mangler.mangle_type(types.CPointer(dtype))
    expected = "P" + expected_str
    assert got == expected


def test_mangling_arg_type_2(dtypes, addrspaces):
    dtype, expected_dtype_str = dtypes
    addrspace, expected_addrspace_str = addrspaces
    got = itanium_mangler.mangle_type(types.CPointer(dtype, addrspace=addrspace))
    expected = "PU" + expected_addrspace_str + expected_dtype_str
    assert got == expected



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from numba.core import types


def assert_no_return(rettype):
    """
    Make sure the type of return is void/None.

    @dppy.kernel does not allow users to return any value and this
    function raises TypeError when users do return something.

    Args:
        rettype: Numba type representing the return value.

    Raises:
        TypeError: Only None and types.void is allowed. In case
            of any other type TypeError is raised.
    """
    if rettype is not None and rettype != types.void:
        msg = "DPPY kernel must have void return type but got {rettype}"
        raise TypeError(msg.format(rettype=rettype))



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import dpctl
import numpy as np
import pytest
from numba import njit

from numba_dppy.tests._helper import dpnp_debug

from .dpnp_skip_test import dpnp_skip_test as skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


@pytest.mark.parametrize(
    "dtype", [np.bool_, np.int32, np.int64, np.float32, np.float64]
)
@pytest.mark.parametrize(
    "shape",
    [(0,), (4,), (2, 3)],
    ids=["(0,)", "(4,)", "(2, 3)"],
)
def test_all(dtype, shape, filter_str):
    if skip_test(filter_str):
        pytest.skip()

    size = 1
    for i in range(len(shape)):
        size *= shape[i]

    for i in range(2 ** size):
        t = i

        a = np.empty(size, dtype=dtype)

        for j in range(size):
            a[j] = 0 if t % 2 == 0 else j + 1
            t = t >> 1

        a = a.reshape(shape)

        def fn(a):
            return np.all(a)

        f = njit(fn)
        with dpctl.device_context(filter_str), dpnp_debug():
            actual = f(a)

        expected = fn(a)
        np.testing.assert_allclose(actual, expected, rtol=1e-3, atol=0)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Provide @vectorize(target="dppy") support."""

import warnings

import dpctl
import numpy as np
from numba.np.ufunc import deviceufunc

import numba_dppy as dppy
from numba_dppy.utils import as_usm_obj, copy_to_numpy_from_usm_obj, has_usm_memory

vectorizer_stager_source = """
def __vectorized_{name}({args}, __out__):
    __tid__ = __dppy__.get_global_id(0)
    if __tid__ < __out__.shape[0]:
        __out__[__tid__] = __core__({argitems})
"""


class DPPYVectorize(deviceufunc.DeviceVectorize):
    def _compile_core(self, sig):
        devfn = dppy.func(sig)(self.pyfunc)
        return devfn, devfn.cres.signature.return_type

    def _get_globals(self, corefn):
        glbl = self.pyfunc.__globals__.copy()
        glbl.update({"__dppy__": dppy, "__core__": corefn})
        return glbl

    def _compile_kernel(self, fnobj, sig):
        return dppy.kernel(sig)(fnobj)

    def build_ufunc(self):
        return DPPYUFuncDispatcher(self.kernelmap)

    @property
    def _kernel_template(self):
        return vectorizer_stager_source


class DPPYUFuncDispatcher(object):
    """
    Invoke the dppy ufunc specialization for the given inputs.
    """

    def __init__(self, types_to_retty_kernels):
        self.functions = types_to_retty_kernels

    def __call__(self, *args, **kws):
        """
        Call the DPPY kernel launching mechanism
        Args:
            *args (np.ndarray): NumPy arrays
            **kws (optional):
                queue (dpctl._sycl_queue.SyclQueue): SYCL queue.
                out (np.ndarray): Output array.
        """
        return DPPYUFuncMechanism.call(self.functions, args, kws)

    def reduce(self, arg, queue=0):
        raise NotImplementedError


class DPPYUFuncMechanism(deviceufunc.UFuncMechanism):
    """
    Mechanism to process Input to a SYCL kernel and launch that kernel
    """

    @classmethod
    def call(cls, typemap, args, kws):
        """
        Perform the entire ufunc call mechanism.
        Args:
            typemap (dict): Signature mapped to kernel.
            args: Arguments to the @vectorize function.
            kws (optional): Optional keywords. Not supported.
        """
        # Handle keywords
        queue = dpctl.get_current_queue()
        out = kws.pop("out", None)

        if kws:
            warnings.warn("unrecognized keywords: %s" % ", ".join(kws))

        # Begin call resolution
        cr = cls(typemap, args)
        args = cr.get_arguments()
        resty, func = cr.get_function()

        outshape = args[0].shape

        # Adjust output value
        if out is not None and cr.is_device_array(out):
            out = cr.as_device_array(out)

        def attempt_ravel(a):
            if cr.SUPPORT_DEVICE_SLICING:
                raise NotImplementedError

            try:
                # Call the `.ravel()` method
                return a.ravel()
            except NotImplementedError:
                # If it is not a device array
                if not cr.is_device_array(a):
                    raise
                # For device array, retry ravel on the host by first
                # copying it back.
                else:
                    hostary = cr.to_host(a, queue).ravel()
                    return cr.to_device(hostary, queue)

        if args[0].ndim > 1:
            args = [attempt_ravel(a) for a in args]

        # Prepare argument on the device
        devarys = []
        any_device = True
        for a in args:
            if cr.is_device_array(a):
                devarys.append(a)
            else:
                dev_a = cr.to_device(a, queue=queue)
                devarys.append(dev_a)

        # Launch
        shape = args[0].shape
        if out is None:
            # No output is provided
            devout = cr.device_array(shape, resty, queue=queue)

            devarys.extend([devout])
            cr.launch(func, shape[0], queue, devarys)

            if any_device:
                # If any of the arguments are on device,
                # Keep output on the device
                return devout.reshape(outshape)
            else:
                # Otherwise, transfer output back to host
                # return devout.copy_to_host().reshape(outshape)
                raise ValueError("copy_to_host() is not yet supported")

        elif cr.is_device_array(out):
            # If output is provided and it is a device array,
            # Return device array
            if out.ndim > 1:
                out = attempt_ravel(out)
            devout = out
            devarys.extend([devout])
            cr.launch(func, shape[0], queue, devarys)
            return devout.reshape(outshape)

        else:
            # If output is provided and it is a host array,
            # Return host array
            assert out.shape == shape
            assert out.dtype == resty
            devout = cr.device_array(shape, resty, queue=queue)
            devarys.extend([devout])
            cr.launch(func, shape[0], queue, devarys)
            return devout.reshape(outshape)

    def as_device_array(self, obj):
        return obj

    def is_device_array(self, obj):
        ret = has_usm_memory(obj)
        return ret is not None

    def is_host_array(self, obj):
        return not self.is_device_array(obj)

    def to_device(self, hostary, queue):
        usm_mem = as_usm_obj(hostary, queue=queue, usm_type="shared")
        usm_backed_ndary = np.ndarray(
            hostary.shape, buffer=usm_mem, dtype=hostary.dtype
        )
        return usm_backed_ndary

    def to_host(self, devary, queue):
        hostary = np.empty(devary.shape, dtype=devary.dtype)
        devary_memview = memoryview(devary)
        devary_memview = devary_memview.cast("B")
        copy_to_numpy_from_usm_obj(devary_memview, hostary)

    def launch(self, func, count, queue, args):
        func[count, dppy.DEFAULT_LOCAL_SIZE](*args)

    def device_array(self, shape, dtype, queue):
        size = np.prod(shape)
        itemsize = dtype.itemsize
        usm_mem = dpctl.memory.MemoryUSMShared(size * itemsize, queue=queue)
        usm_backed_ndary = np.ndarray(shape, buffer=usm_mem, dtype=dtype)
        return usm_backed_ndary

    def broadcast_device(self, ary, shape):
        raise NotImplementedError("device broadcast_device NIY")



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from llvmlite import binding as ll
from llvmlite.llvmpy import core as lc
from numba.core import utils
from numba.core.codegen import CPUCodegen, CPUCodeLibrary

from numba_dppy import config

SPIR_TRIPLE = {32: " spir-unknown-unknown", 64: "spir64-unknown-unknown"}

SPIR_DATA_LAYOUT = {
    32: (
        "e-p:32:32-i64:64-v16:16-v24:32-v32:32-v48:64-v96:128-v192:256-v256:"
        "256-v512:512-v1024:1024"
    ),
    64: (
        "e-i64:64-v16:16-v24:32-v32:32-v48:64-v96:128-v192:256-v256:256-"
        "v512:512-v1024:1024"
    ),
}


class SPIRVCodeLibrary(CPUCodeLibrary):
    def _optimize_functions(self, ll_module):
        pass

    def _optimize_final_module(self):
        # Run some lightweight optimization to simplify the module.
        pmb = ll.PassManagerBuilder()

        # Make optimization level depending on config.OPT variable
        pmb.opt_level = config.OPT

        pmb.disable_unit_at_a_time = False
        pmb.disable_unroll_loops = True
        pmb.loop_vectorize = False
        pmb.slp_vectorize = False

        pm = ll.ModulePassManager()
        pmb.populate(pm)
        pm.run(self._final_module)

    def _finalize_specific(self):
        # Fix global naming
        for gv in self._final_module.global_variables:
            if "." in gv.name:
                gv.name = gv.name.replace(".", "_")

    def get_asm_str(self):
        # Return nothing: we can only dump assembler code when it is later
        # generated (in numba_dppy.compiler).
        return None


class JITSPIRVCodegen(CPUCodegen):
    """
    This codegen implementation generates optimized SPIR 2.0
    """

    _library_class = SPIRVCodeLibrary

    def _init(self, llvm_module):
        assert list(llvm_module.global_variables) == [], "Module isn't empty"
        self._data_layout = SPIR_DATA_LAYOUT[utils.MACHINE_BITS]
        self._target_data = ll.create_target_data(self._data_layout)

    def _create_empty_module(self, name):
        ir_module = lc.Module(name)
        ir_module.triple = SPIR_TRIPLE[utils.MACHINE_BITS]
        if self._data_layout:
            ir_module.data_layout = self._data_layout
        return ir_module

    def _module_pass_manager(self):
        raise NotImplementedError

    def _function_pass_manager(self, llvm_module):
        raise NotImplementedError

    def _add_module(self, module):
        pass



################################################################################
#                                 Numba-DPPY
#
# Copyright 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
################################################################################

import dpctl
import numpy as np
import pytest
from numba import njit

import numba_dppy as dppy
from numba_dppy.tests._helper import dpnp_debug

from .dpnp_skip_test import dpnp_skip_test as skip_test

list_of_filter_strs = [
    "opencl:gpu:0",
    "level_zero:gpu:0",
    "opencl:cpu:0",
]


@pytest.fixture(params=list_of_filter_strs)
def filter_str(request):
    return request.param


@pytest.mark.parametrize(
    "arr",
    [[1, 2, 3, 4], [1, 2, 3, 4, 5, 6, 7, 8, 9]],
    ids=["[1, 2, 3, 4]", "[1, 2, 3, 4, 5, 6, 7, 8, 9]"],
)
def test_repeat(filter_str, arr):
    if skip_test(filter_str):
        pytest.skip()

    a = np.array(arr)
    repeats = 2

    def fn(a, repeats):
        return np.repeat(a, repeats)

    f = njit(fn)
    with dpctl.device_context(filter_str), dpnp_debug():
        actual = f(a, repeats)

    expected = fn(a, repeats)
    np.testing.assert_allclose(actual, expected, rtol=1e-3, atol=0)



# Copyright 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum, auto


class sycl_memory_order(Enum):
    """
    An enumeration of SYCL memory order. For more details please refer to
    SYCL 2020 specification, section 3.8.3.1
    (https://www.khronos.org/registry/SYCL/specs/sycl-2020/html/sycl-2020.html#_memory_ordering).

    For DPCPP implementation please refer:
    https://github.com/intel/llvm/blob/sycl-nightly/20210507/sycl/include/CL/sycl/ONEAPI/atomic_enums.hpp#L25

    =====================   ============
    Memory Order            Enum value
    =====================   ============
    relaxed                 0
    acquire                 1
    __consume_unsupported   2
    release                 3
    acq_rel                 4
    seq_cst                 5
    =====================   ============
    """

    relaxed = auto()
    acquire = auto()
    __consume_unsupported = auto()
    release = auto()
    acq_rel = auto()
    seq_cst = auto()


class sycl_memory_scope(Enum):
    """
    An enumeration of SYCL memory scope. For more details please refer to
    SYCL 2020 specification, section 3.8.3.2
    (https://www.khronos.org/registry/SYCL/specs/sycl-2020/html/sycl-2020.html#_memory_scope).

    For DPCPP implementation please refer:
    https://github.com/intel/llvm/blob/sycl-nightly/20210507/sycl/include/CL/sycl/ONEAPI/atomic_enums.hpp#L45

    ===============  ============
    Memory Scope     Enum value
    ===============  ============
    work_item        0
    sub_group        1
    work_group       2
    device           3
    system           4
    ===============  ============
    """

    work_item = auto()
    sub_group = auto()
    work_group = auto()
    device = auto()
    system = auto()


class _spv_scope(Enum):
    """
    An enumeration of SPIRV scope.

    For DPCPP implementation please refer:
    https://github.com/intel/llvm/blob/sycl-nightly/20210507/sycl/include/CL/__spirv/spirv_types.hpp#L24
    """

    CrossDevice = 0
    Device = 1
    Workgroup = 2
    Subgroup = 3
    Invocation = 4


class _spv_memory_semantics_mask(Enum):
    """
    An enumeration of SPIRV memory semantics mask.

    For DPCPP implementation please refer:
    https://github.com/intel/llvm/blob/sycl-nightly/20210507/sycl/include/CL/__spirv/spirv_types.hpp#L81
    """

    NONE = 0x0
    Acquire = 0x2
    Release = 0x4
    AcquireRelease = 0x8
    SequentiallyConsistent = 0x10
    UniformMemory = 0x40
    SubgroupMemory = 0x80
    WorkgroupMemory = 0x100
    CrossWorkgroupMemory = 0x200
    AtomicCounterMemory = 0x400
    ImageMemory = 0x800


def get_memory_semantics_mask(memory_order):
    """
    This function translates SYCL memory order to SPIRV memory semantics mask.

    For DPCPP implementation please refer:
    https://github.com/intel/llvm/blob/sycl-nightly/20210507/sycl/include/CL/sycl/detail/spirv.hpp#L220
    """

    spv_order = _spv_memory_semantics_mask.NONE.value
    if memory_order == sycl_memory_order.relaxed:
        spv_order = _spv_memory_semantics_mask.NONE.value
    elif memory_order == sycl_memory_order.__consume_unsupported:
        pass
    elif memory_order == sycl_memory_order.acquire:
        spv_order = _spv_memory_semantics_mask.Acquire.value
    elif memory_order == sycl_memory_order.release:
        spv_order = _spv_memory_semantics_mask.Release.value
    elif memory_order == sycl_memory_order.acq_rel:
        spv_order = _spv_memory_semantics_mask.AcquireRelease.value
    elif memory_order == sycl_memory_order.seq_cst:
        spv_order = _spv_memory_semantics_mask.SequentiallyConsistent.value
    else:
        raise ValueError("Invalid memory order provided")

    return (
        spv_order
        | _spv_memory_semantics_mask.SubgroupMemory.value
        | _spv_memory_semantics_mask.WorkgroupMemory.value
        | _spv_memory_semantics_mask.CrossWorkgroupMemory.value
    )


def get_scope(memory_scope):
    """
    This function translates SYCL memory scope to SPIRV scope.

    For DPCPP implementation please refer:
    https://github.com/intel/llvm/blob/sycl-nightly/20210507/sycl/include/CL/sycl/detail/spirv.hpp#L247
    """

    if memory_scope == sycl_memory_scope.work_item:
        return _spv_scope.Invocation.value
    elif memory_scope == sycl_memory_scope.sub_group:
        return _spv_scope.Subgroup.value
    elif memory_scope == sycl_memory_scope.work_group:
        return _spv_scope.Workgroup.value
    elif memory_scope == sycl_memory_scope.device:
        return _spv_scope.Device.value
    elif memory_scope == sycl_memory_scope.system:
        return _spv_scope.CrossDevice.value
    else:
        raise ValueError("Invalid memory scope provided")
